import numpy as np
from . import DataBase
from astropy.coordinates import SkyCoord, EarthLocation
from astropy import units as u
from astropy.time import Time
import sqlite3, time, warnings, os
import numpy as np
import multiprocessing as mp
from astropy.coordinates.angle_utilities import angular_separation
from .aleph_utils import *
from astropy.table import QTable, Table, hstack

def getint(s):
    try: return int(s)
    except: return np.nan

PACKAGE_PATH = os.path.abspath(os.path.dirname(__file__))
IS_SQL_DB = os.path.exists(PACKAGE_PATH+'/aleph_states.db')

class AstEph:
    def __init__(self, service='Lowell', open_db=IS_SQL_DB, dbfile=PACKAGE_PATH+'/aleph_states.db'):
        """
        Initializing 'Query' enviroment. Orbital parameters are loaded
        and the sql database with asteroid's states is connected.

        Parameters:
                    
        """
        if (service=='MPC' and open_db==True):
            w = "You are using the 'MPC' survey AND opening the 'dbfile'='%s'" %dbfile
            w+= ". Only 'query_2b_cat()' and 'query_nb_cat()' will use 'MPCORB.DAT' data."
            warnings.warn(w)
        self.service = service
        alephs = DataBase.OrbParams(service)
        self.asts = alephs.asts
        self.ref_epochs = alephs.dicc_epochs_M
        self.ref_jdtdb = {t:self.ref_epochs[t].tdb.jd for t in self.ref_epochs}
        nums = []; nms = []; hs = []; gs = []
        for ast in self.asts:
            nums.append(getint(ast['num'])); nms.append(ast['name'].strip())
            hs.append(ast['H']); gs.append(ast['G'])
        self.numbers = np.array(nums)
        self.names = np.array(nms)
        self.H = np.array(hs)
        self.G = np.array(gs)
        del alephs, nums, nms, hs, gs
        
        # Opening database
        if open_db:
            self.connection = sqlite3.connect(dbfile)
            self.cursor = self.connection.cursor()
            
            self.cursor.execute('SELECT name,number,H,G FROM asts_info')
            info = np.array(self.cursor.fetchall(), dtype=str)
            self.numbers_db = np.array([getint(num) for num in info[:,1]])
            self.names_db = np.array([nm.strip() for nm in info[:,0]])
            self.H_db = info[:,2].astype(float); self.G_db = info[:,3].astype(float)


import numpy as np
from . import DataBase
from astropy.coordinates import SkyCoord, EarthLocation
from astropy import units as u
from astropy.time import Time
import sqlite3, time, warnings, os
import numpy as np
import multiprocessing as mp
from astropy.coordinates.angle_utilities import angular_separation
from .aleph_utils import *
from astropy.table import QTable, Table, hstack
import pandas as pd

def getint(s):
    try: return int(s)
    except: return np.nan

PACKAGE_PATH = os.path.abspath(os.path.dirname(__file__))
IS_SQL_DB = os.path.exists(PACKAGE_PATH+'/aleph_states.db')

class Query:
    def __init__(self, service='MPC', open_db=IS_SQL_DB, dbfile=PACKAGE_PATH+'/aleph_states.db', **kwargs):
        """
        Initializing 'Query' enviroment. Orbital parameters are loaded
        and the sql database with asteroid's states is connected.

        Parameters:
                    service: Sets origin of orbital parameters. Can be 'Lowell' or 'MPC'.
                             Default to 'Lowell'.
        Other Parameters:
                    open_db
                    dbfile
                    filename: Str. Specifies file (and path) with orbital parameters to
                              read. This file must be in Lowell's or MPC's format.
        Attributes:
                    service: 'Lowell' or 'MPC'
                    asts: List of dictionaries containing all asteroids' information.
                    ref_epochs
                    ref_jdtdb
                    numbers
                    names
                    H
                    G
                    connection
                    cursor
                    asts_db
                    numbers_db
                    names_db
                    H_db
                    G_db
        """
        if (service=='MPC' and open_db==True):
            w = "You are using the 'MPC' survey AND opening the 'dbfile'='%s'" %dbfile
            w+= ". Only 'query_2b_cat()' and 'query_nb_cat()' will use 'MPCORB.DAT' data."
            warnings.warn(w)
        self.service = service
        if 'filename' in kwargs: alephs = DataBase.OrbParams(service, filename=kwargs['filename'])
        else: alephs = DataBase.OrbParams(service)
        self.asts = alephs.asts
        self.ref_epochs = alephs.dicc_epochs_M
        self.ref_jdtdb = {t:self.ref_epochs[t].tdb.jd for t in self.ref_epochs}
        nums = []; nms = []; hs = []; gs = []
        for ast in self.asts:
            nums.append(getint(ast['num'])); nms.append(ast['name'].strip())
            hs.append(ast['H']); gs.append(ast['G'])
        self.numbers = np.array(nums)
        self.names = np.array(nms)
        self.H = np.array(hs)
        self.G = np.array(gs)
        del alephs, nums, nms, hs, gs
        
        # Opening database
        if open_db:
            self.connection = sqlite3.connect(dbfile)
            self.cursor = self.connection.cursor()
            
            self.asts_db = pd.read_sql('SELECT * FROM asts_info', self.connection, index_col='name')
            
            self.numbers_db = np.array([getint(num) for num in self.asts_db.number.values])
            self.names_db = self.asts_db.index.values; self.H_db = self.asts_db.H.values; self.G_db = self.asts_db.G.values
    
    #def query(self, field_center, radius, **kwargs):
    #    """
    #    Query for asteroides visible in a field.
    #    
    #    Parameters:
    #                field_center: astropy.coordinates.SkyCoord element with center coordinate of field.
    #                radius: astropy.units.Quantity with radius of field.
    #                
    #    Optional Parameters:
    #                epoch: astropy.time.Time with epoch for query. Default to 'now'.
    #                observer: astropy.coordinates.EarthLocation with observer's position. Default to geocenter.
     #               njobs: Number of parallel processes to calculate asteroids' apparent positions. Default: 1.
     #               method: 'db', '2B' or 'NB'. Default 'db'.
     #                      'db': Queries database for bodies near field and integrates them.
     #                      '2B': Calculates all bodies positions using 2-Bodies equations and retrieves bodies in
     #                            requested field.
     #                      'NB': Calculates all bodies positions using N-Bodies integration (Sun, planets, Pluto and 
     #                            massless asteroids) and retrieves bodies in requested field.
     #               db_query_radius: Radius to query database. Only usefull if 'method'='db'.
     #                                It should be greater than 'radius'. Default: 5 degs.
     #               ast_names: Asteroids' name from Query.asts. Ephemerides will be calculated only for those
     #                          asteroids.
     #               
     #   Returns:
     #            astropy.table with computed ephemerides.
     #             column 'number': Asteroid's number (from MPC).
     #             column 'name': Asteroid's name (from MPC).
     #             column 'ra': Asteroid's apparent right ascension (as seen by the observer).
     #             column 'dec': Asteroid's apparent declination (as seen by the observer).
     #             column 'x': Asteroid's heliocentric cartesian 'x' coordinate at light-emission time.
     #             column 'y': Asteroid's heliocentric cartesian 'y' coordinate at light-emission time.
     #             column 'z': Asteroid's heliocentric cartesian 'z' coordinate at light-emission time.
     #             column 'vx': Asteroid's heliocentric cartesian 'vx' velocity at light-emission time.
     #             column 'vy': Asteroid's heliocentric cartesian 'vy' velocity at light-emission time.
     #             column 'vz': Asteroid's heliocentric cartesian 'vz' velocity at light-emission time.
     #             column 'lightdist': Distance travel by light from asteroid's coordinate until observer.
     #   """
     #   # Selecting method
     #   if not 'method' in kwargs: return self.query_db(field_center, radius, **kwargs)
     #   else:
     #       if kwargs['method']=='db': return self.query_db(field_center, radius, **kwargs)
     #       elif kwargs['method']=='2B': return self.query_2b(field_center, radius, **kwargs)
     #       elif kwargs['method']=='NB': return self.query_nb(field_center, radius, **kwargs)
     #       else: raise ValueError("'method' must be one of 'db', '2B' or 'NB'")
    
    
    def query_2b_cat(self, field_center, radius, **kwargs):
        """
        Query for asteroides visible in a field solving the 2-body equation from the asteroids'
        orbital parameters stored in an 'astorb' or an 'MPCORB' catalogue.
        
        Parameters:
                    field_center: astropy.coordinates.SkyCoord element with center coordinate of field.
                    radius: astropy.units.Quantity with radius of field.
                    
        Optional Parameters:
                    epoch: astropy.time.Time with epoch for query. Default to 'now'.
                    observer: astropy.coordinates.EarthLocation with observer's position. Default to geocenter.
                    njobs: Number of parallel processes to calculate asteroids' apparent positions. Default: 1.
                    ast_idxs: Asteroids' indexes (starting from 0) from Query.asts. Ephemerides will be
                              calculated only for those asteroids.
                    full: bool. If True, ephemeris are calculated for all asteroids (or those specified in
                          'ast_idxs') before selectig those in the requested field. If False, instantaneous
                          coordinates for all asteroids (or those specified in 'ast_idxs') are calculated first
                          and then apparent coordinates are calculated only for those bodies with instantaneos
                          coordinates inside a wider field. Default: False.
                    confidence_radius: Only when 'full=False'. Specifies the radius around requested 'field_center'
                                       to select the asteroids to calculate their apparent ephemeris. Default:
                                       radius+(3 degrees).

                    
        Returns:
                 astropy.table with computed ephemerides.
                  column 'number': Asteroid's number (from MPC).
                  column 'name': Asteroid's name (from MPC).
                  column 'ra': Asteroid's apparent right ascension (as seen by the observer).
                  column 'dec': Asteroid's apparent declination (as seen by the observer).
                  column 'x': Asteroid's heliocentric cartesian 'x' coordinate at light-emission time.
                  column 'y': Asteroid's heliocentric cartesian 'y' coordinate at light-emission time.
                  column 'z': Asteroid's heliocentric cartesian 'z' coordinate at light-emission time.
                  column 'vx': Asteroid's heliocentric cartesian 'vx' velocity at light-emission time.
                  column 'vy': Asteroid's heliocentric cartesian 'vy' velocity at light-emission time.
                  column 'vz': Asteroid's heliocentric cartesian 'vz' velocity at light-emission time.
                  column 'lightdist': Distance travel by light from asteroid's coordinate until observer.
        """
        # Getting all necessary parameters
        if 'observer' in kwargs: observer = kwargs['observer']
        else: observer = EarthLocation.from_geocentric(0,0,0, unit='m')
        if 'epoch' in kwargs: epoch = kwargs['epoch']
        else: epoch = Time.now()
        if 'njobs' in kwargs: njobs = kwargs['njobs']
        else: njobs = 1
        if 'ast_idxs' in kwargs: ast_idxs = np.array(kwargs['ast_idxs'])
        else: ast_idxs = np.arange(len(self.asts))
        if 'full' in kwargs: full = kwargs['full']
        else: full = False
        if 'confidence_radius' in kwargs: confidence_radius = kwargs['confidence_radius'].to('deg').value
        else: confidence_radius = radius.to('deg').value + 3
        [geopos,geovel] = observer.get_gcrs_posvel(epoch)
        obscoords = SkyCoord(geopos, frame='gcrs', obstime=epoch)
        obsx = obscoords.hcrs.cartesian.x.to('au').value
        obsy = obscoords.hcrs.cartesian.y.to('au').value
        obsz = obscoords.hcrs.cartesian.z.to('au').value
        tq = epoch.tdb.jd

        if not full:
            # Instantaneous states are calculated for all asteroids.
            states = []
            for idx in ast_idxs:
                ast = self.asts[idx]
                tr = self.ref_jdtdb[ast['epoch']]
                r = pycutils.state_equatorial_heliocentric(tq, tr, ast['a'],ast['e'],ast['incl'],ast['Node'],ast['Arg_Peri'],ast['M_epoch'])
                states.append(r)
            [xh ,yh, zh] = np.array(states)[:,:3].T
            # Instantaneous ephemerides are calculated for all asteroids.
            xe = xh-obsx; ye = yh-obsy; ze = zh-obsz
            ra_geo  = np.arctan2( ye, xe ); dec_geo = np.arctan2( ze, np.sqrt(xe*xe+ye*ye) )
            dists = angular_separation(field_center.ra.rad, field_center.dec.rad, ra_geo, dec_geo)
            inconffield = dists<(confidence_radius*np.pi/180)
            # Ephemeris are calculated for those bodies inside confidence field
            ast_idxs = ast_idxs[inconffield]
        
        # Apparent coordinates are calculated
        if njobs==1:
            coords = []
            for idx in ast_idxs:
                ast = self.asts[idx]
                tr = self.ref_jdtdb[ast['epoch']]
                args =  tq, obsx, obsy, obsz, tr, ast['a'], ast['e'], ast['incl'], ast['Node'], ast['Arg_Peri'], ast['M_epoch']
                coords.append(pycutils.apparent_coords_equatorial_heliocentric(*args))
        else:
            args = []
            for idx in ast_idxs:
                ast = self.asts[idx]
                tr = self.ref_jdtdb[ast['epoch']]
                args.append((tq, obsx, obsy, obsz, tr, ast['a'], ast['e'], ast['incl'], ast['Node'], ast['Arg_Peri'], ast['M_epoch']))
            pool = mp.Pool(processes=njobs)
            coords = pool.starmap(pycutils.apparent_coords_equatorial_heliocentric, args)
            pool.close()

        # Getting final coordinates
        coords = np.array(coords); ast_idxs_mask = tuple(ast_idxs),
        #[ra,dec,x,y,z,vx,vy,vz,geodist] = coords.T # deg, au
        ra = coords[:,0]; dec = coords[:,1]
        dists = angular_separation(field_center.ra, field_center.dec, ra*u.deg, dec*u.deg)
        infield = dists.to('rad').value<radius.to('rad').value
        coords = coords[infield]; nums = self.numbers[ast_idxs_mask][infield]; nms = self.names[ast_idxs_mask][infield]
        hs = self.H[ast_idxs_mask][infield]; gs = self.G[ast_idxs_mask][infield]
        
        [x ,y, z] = coords[:,2:5].T; geodist = coords[:,8] # au
        xe, ye, ze = x-obsx, y-obsy, z-obsz
        ast_geo = [-xe, -ye, -ze]   # Obs vector from ast
        ast_sun = [-x, -y, -z]      # Sun vector from ast
        alpha = angle_between_vectors(ast_geo, ast_sun)  # Phase angle (in rads)
        phase_func1 = np.exp(-3.33*np.tan(alpha/2)**0.36)
        phase_func2 = np.exp(-1.87*np.tan(alpha/2)**1.22)
        phase_func = (1-gs)*phase_func1 + gs*phase_func2
        Vs = hs + 5*np.log10(geodist*np.sqrt(x*x+y*y+z*z)) - 2.5*np.log10(phase_func)
        #return coords, nums, nms
        
        geo_ast = [xe, ye, ze]            # Obs vector from ast
        geo_sun = [-obsx, -obsy, -obsz]   # Sun vector from ast
        elong = angle_between_vectors(geo_ast, geo_sun)  # Elongation angle (in rads)
        to_deg = 180/np.pi
        
        t0 = Table((nums, nms), names=('number','name'))
        cnames=('ra','dec','x','y','z','vx','vy','vz','lightdist')
        us = (u.deg, u.deg, u.au, u.au, u.au, u.au/u.day, u.au/u.day, u.au/u.day, u.au)
        cols = [column*us[i] for i, column in enumerate(coords.T)]
        t = QTable(cols, names=cnames)
        t2= QTable([alpha*to_deg*u.deg, elong*to_deg*u.deg, Vs], names=['phase', 'elongation', 'V'])
        return hstack([t0, t, t2])

    
    def query_nb_cat(self, field_center, radius, **kwargs):
        """
        Query for asteroides visible in a field. Each asteroids is initialized using its orbital parameters
        and then it is integrated until requested epoch using an n-body simulation including the Sun,
        the planets, Pluto and the asteroid as a massless particle.
        
        Parameters:
                    field_center: astropy.coordinates.SkyCoord element with center coordinate of field.
                    radius: astropy.units.Quantity with radius of field.
                    
        Optional Parameters:
                    epoch: astropy.time.Time with epoch for query. Default to 'now'.
                    observer: astropy.coordinates.EarthLocation with observer's position. Default to geocenter.
                    njobs: Number of parallel processes to calculate asteroids' apparent positions. Default: 1.
                    ast_idxs: Asteroids' indexes (starting from 0) from Query.asts. Ephemerides will be
                              calculated only for those asteroids.
                    
        Returns:
                 astropy.table with results.
                  column 'number': Asteroid's number (from MPC).
                  column 'name': Asteroid's name (from MPC).
                  column 'ra': Asteroid's apparent right ascension (as seen by the observer).
                  column 'dec': Asteroid's apparent declination (as seen by the observer).
                  column 'x': Asteroid's heliocentric cartesian 'x' coordinate at light-emission time.
                  column 'y': Asteroid's heliocentric cartesian 'y' coordinate at light-emission time.
                  column 'z': Asteroid's heliocentric cartesian 'z' coordinate at light-emission time.
                  column 'vx': Asteroid's heliocentric cartesian 'vx' velocity at light-emission time.
                  column 'vy': Asteroid's heliocentric cartesian 'vy' velocity at light-emission time.
                  column 'vz': Asteroid's heliocentric cartesian 'vz' velocity at light-emission time.
                  column 'lightdist': Distance travel by light from asteroid's coordinate until observer.
        """
        # Getting all necessary parameters
        if 'observer' in kwargs: observer = kwargs['observer']
        else: observer = EarthLocation.from_geocentric(0,0,0, unit='m')
        if 'epoch' in kwargs: epoch = kwargs['epoch']
        else: epoch = Time.now()
        if 'njobs' in kwargs: njobs = kwargs['njobs']
        else: njobs = 1
        if 'ast_idxs' in kwargs: ast_idxs = kwargs['ast_idxs']
        else: ast_idxs = np.arange(len(self.asts))
        [geopos,geovel] = observer.get_gcrs_posvel(epoch)
        obscoords = SkyCoord(geopos, frame='gcrs', obstime=epoch)
        obsx = obscoords.hcrs.cartesian.x.to('au').value
        obsy = obscoords.hcrs.cartesian.y.to('au').value
        obsz = obscoords.hcrs.cartesian.z.to('au').value
        tq = epoch.tdb.jd

        # Getting Solar System's states in reference epochs
        set_ref_epochs = set([self.asts[idx]['epoch'] for idx in ast_idxs])
        dicc_ss_states = {epoch:get_SS_states(self.ref_epochs[epoch], frame='helio') for epoch in set_ref_epochs}
        
        if njobs==1:
            coords = []
            for idx in ast_idxs:
                ast = self.asts[idx]
                tr = self.ref_jdtdb[ast['epoch']]; dt = tq-tr
                ss_states = dicc_ss_states[ast['epoch']]
                args =  tr, tr, ast['a'], ast['e'], ast['incl'], ast['Node'], ast['Arg_Peri'], ast['M_epoch']
                state_ast=pycutils.state_equatorial_heliocentric(*args)
                coords.append(get_apparent_coordinates(state_ast, ss_states, obsx, obsy, obsz, dt))
        else:
            arguments = []
            for idx in ast_idxs:
                ast = self.asts[idx]
                tr = self.ref_jdtdb[ast['epoch']]; dt = tq-tr
                ss_states = dicc_ss_states[ast['epoch']]
                args =  tr, tr, ast['a'], ast['e'], ast['incl'], ast['Node'], ast['Arg_Peri'], ast['M_epoch']
                state = pycutils.state_equatorial_heliocentric(*args)
                arguments.append((state, ss_states, obsx, obsy, obsz, dt))
            pool = mp.Pool(processes=njobs)
            coords = pool.starmap(get_apparent_coordinates, arguments)
            pool.close()
            
        # Getting final coordinates
        coords = np.array(coords); ast_idxs_mask = tuple(ast_idxs),
        #[ra_geo, dec_geo, x ,y, z, vx,vy, vz, geodist] = coords.T # rad, au, au/d
        ra_geo = coords[:,0]; dec_geo = coords[:,1]
        dists = angular_separation(field_center.ra.rad, field_center.dec.rad, ra_geo, dec_geo)
        infield = dists<radius.to('rad').value
        coords = coords[infield]; coords[:,0] = coords[:,0]*180/np.pi; coords[:,1] = coords[:,1]*180/np.pi
        nums = self.numbers[ast_idxs_mask][infield]; nms = self.names[ast_idxs_mask][infield]
        hs = self.H[ast_idxs_mask][infield]; gs = self.G[ast_idxs_mask][infield]
        
        [x ,y, z] = coords[:,2:5].T; geodist = coords[:,8] # au
        xe, ye, ze = x-obsx, y-obsy, z-obsz
        ast_geo = [-xe, -ye, -ze]   # Obs vector from ast
        ast_sun = [-x, -y, -z]      # Sun vector from ast
        alpha = angle_between_vectors(ast_geo, ast_sun)
        phase_func1 = np.exp(-3.33*np.tan(alpha/2)**0.36)
        phase_func2 = np.exp(-1.87*np.tan(alpha/2)**1.22)
        phase_func = (1-gs)*phase_func1 + gs*phase_func2
        Vs = hs + 5*np.log10(geodist*np.sqrt(x*x+y*y+z*z)) - 2.5*np.log10(phase_func)
        #return coords, nums, nms
        
        geo_ast = [xe, ye, ze]            # Obs vector from ast
        geo_sun = [-obsx, -obsy, -obsz]   # Sun vector from ast
        elong = angle_between_vectors(geo_ast, geo_sun)  # Elongation angle (in rads)
        to_deg = 180/np.pi
        
        t0 = Table((nums, nms), names=('number','name'))
        cnames=('ra','dec','x','y','z','vx','vy','vz','lightdist')
        us = (u.deg, u.deg, u.au, u.au, u.au, u.au/u.day, u.au/u.day, u.au/u.day, u.au)
        cols = [column*us[i] for i, column in enumerate(coords.T)]
        t = QTable(cols, names=cnames)
        t2= QTable([alpha*to_deg*u.deg, elong*to_deg*u.deg, Vs], names=['phase', 'elongation', 'V'])
        return hstack([t0, t, t2]) 

    def query_mixed_cat(self, field_center, radius, **kwargs):
        """
        Query for asteroides visible in a field. Asteroids ephemeris are pre-calculated using
        their orbital parameters. The ephemeris of tose asteroids that are within a 'confidence
        radius' (larger than the query's radius) are then integrated to increase their accuracy.
        Ephemeris integrations consider asteroids as massless particles under the influence of
        the Sun, the planets and Pluto.
    
        Parameters:
                    field_center: astropy.coordinates.SkyCoord element with center coordinate of field.
                    radius: astropy.units.Quantity with radius of field.
                    
        Optional Parameters:
                    epoch: astropy.time.Time with epoch for query. Default to 'now'.
                    observer: astropy.coordinates.EarthLocation with observer's position. Default to geocenter.
                    njobs: Number of parallel processes to calculate asteroids' apparent positions. Default: 1.
                    confidence_radius: The 2-Body ephemeris of the asteroids within this radius (around
                                       'field_center') are selected to be integrated. It should be greater
                                       than 'radius'. Default: 5 degs.
                    ast_idxs: Asteroids' indexes (starting from 0) from Query.asts. Ephemerides will be
                              calculated only for those asteroids. It works as a mask of the asteroids data.
                    
        Returns:
                 astropy.table with computed ephemerides:
                  column 'number': Asteroid's number (from MPC).
                  column 'name': Asteroid's name (from MPC).
                  column 'ra': Asteroid's apparent right ascension (as seen by the observer).
                  column 'dec': Asteroid's apparent declination (as seen by the observer).
                  column 'x': Asteroid's heliocentric cartesian 'x' coordinate at light-emission time.
                  column 'y': Asteroid's heliocentric cartesian 'y' coordinate at light-emission time.
                  column 'z': Asteroid's heliocentric cartesian 'z' coordinate at light-emission time.
                  column 'vx': Asteroid's heliocentric cartesian 'vx' velocity at light-emission time.
                  column 'vy': Asteroid's heliocentric cartesian 'vy' velocity at light-emission time.
                  column 'vz': Asteroid's heliocentric cartesian 'vz' velocity at light-emission time.
                  column 'lightdist': Distance travel by light from asteroid's coordinate until observer.
        """
        # Getting all necessary parameters
        if 'observer' in kwargs: observer = kwargs['observer']
        else: observer = EarthLocation.from_geocentric(0,0,0, unit='m')
        if 'epoch' in kwargs: epoch = kwargs['epoch']
        else: epoch = Time.now()
        if 'njobs' in kwargs: njobs = kwargs['njobs']
        else: njobs = 1
        if 'ast_idxs' in kwargs: ast_idxs = np.array(kwargs['ast_idxs'])
        else: ast_idxs = np.arange(len(self.asts))
        if 'confidence_radius' in kwargs: confidence_radius = kwargs['confidence_radius'].to('deg').value
        else: confidence_radius = radius.to('deg').value + 3
        [geopos,geovel] = observer.get_gcrs_posvel(epoch)
        obscoords = SkyCoord(geopos, frame='gcrs', obstime=epoch)
        obsx = obscoords.hcrs.cartesian.x.to('au').value
        obsy = obscoords.hcrs.cartesian.y.to('au').value
        obsz = obscoords.hcrs.cartesian.z.to('au').value
        tq = epoch.tdb.jd

        # Instantaneous states are calculated for all asteroids.
        states = []
        for idx in ast_idxs:
            ast = self.asts[idx]
            tr = self.ref_jdtdb[ast['epoch']]
            r = pycutils.state_equatorial_heliocentric(tq, tr, ast['a'],ast['e'],ast['incl'],ast['Node'],ast['Arg_Peri'],ast['M_epoch'])
            states.append(r)
        [xh ,yh, zh] = np.array(states)[:,:3].T
        # Instantaneous ephemerides are calculated for all asteroids.
        xe = xh-obsx; ye = yh-obsy; ze = zh-obsz
        ra_geo  = np.arctan2( ye, xe ); dec_geo = np.arctan2( ze, np.sqrt(xe*xe+ye*ye) )
        dists = angular_separation(field_center.ra.rad, field_center.dec.rad, ra_geo, dec_geo)
        inconffield = dists<(confidence_radius*np.pi/180)
        # Ephemeris are calculated for those bodies inside confidence field
        ast_idxs = ast_idxs[inconffield]
        
        # Getting Solar System's states in pertinent reference epochs
        set_ref_epochs = set([self.asts[idx]['epoch'] for idx in ast_idxs])
        dicc_ss_states = {epoch:get_SS_states(self.ref_epochs[epoch], frame='helio') for epoch in set_ref_epochs}
        
        # Apparent coordinates are calculated integrating the N-body problem
        if njobs==1:
            coords = []
            for idx in ast_idxs:
                ast = self.asts[idx]
                tr = self.ref_jdtdb[ast['epoch']]; dt = tq-tr
                ss_states = dicc_ss_states[ast['epoch']]
                args =  tr, tr, ast['a'], ast['e'], ast['incl'], ast['Node'], ast['Arg_Peri'], ast['M_epoch']
                state_ast=pycutils.state_equatorial_heliocentric(*args)
                coords.append(get_apparent_coordinates(state_ast, ss_states, obsx, obsy, obsz, dt))
        else:
            arguments = []
            for idx in ast_idxs:
                ast = self.asts[idx]
                tr = self.ref_jdtdb[ast['epoch']]; dt = tq-tr
                ss_states = dicc_ss_states[ast['epoch']]
                args =  tr, tr, ast['a'], ast['e'], ast['incl'], ast['Node'], ast['Arg_Peri'], ast['M_epoch']
                state = pycutils.state_equatorial_heliocentric(*args)
                arguments.append((state, ss_states, obsx, obsy, obsz, dt))
            pool = mp.Pool(processes=njobs)
            coords = pool.starmap(get_apparent_coordinates, arguments)
            pool.close()

        # Getting final coordinates
        coords = np.array(coords); ast_idxs_mask = tuple(ast_idxs),
        #[ra,dec,x,y,z,vx,vy,vz,geodist] = coords.T # deg, au
        ra = coords[:,0]; dec = coords[:,1]
        dists = angular_separation(field_center.ra.rad, field_center.dec.rad, ra, dec)
        infield = dists<radius.to('rad').value
        coords = coords[infield]
        coords[:,0] = coords[:,0]*180/np.pi; coords[:,1] = coords[:,1]*180/np.pi
        nums = self.numbers[ast_idxs_mask][infield]
        nms = self.names[ast_idxs_mask][infield]
        hs = self.H[ast_idxs_mask][infield]
        gs = self.G[ast_idxs_mask][infield]
        
        [x ,y, z] = coords[:,2:5].T; geodist = coords[:,8] # au
        xe, ye, ze = x-obsx, y-obsy, z-obsz
        ast_geo = [-xe, -ye, -ze]   # Obs vector from ast
        ast_sun = [-x, -y, -z]      # Sun vector from ast
        alpha = angle_between_vectors(ast_geo, ast_sun)
        phase_func1 = np.exp(-3.33*np.tan(alpha/2)**0.36)
        phase_func2 = np.exp(-1.87*np.tan(alpha/2)**1.22)
        phase_func = (1-gs)*phase_func1 + gs*phase_func2
        Vs = hs + 5*np.log10(geodist*np.sqrt(x*x+y*y+z*z)) - 2.5*np.log10(phase_func)
        #return coords, nums, nms
        
        geo_ast = [xe, ye, ze]         # Obs vector from ast
        geo_sun = [-obsx, -obsy, -obsz]   # Sun vector from ast
        elong = angle_between_vectors(geo_ast, geo_sun)  # Elongation angle (in rads)
        to_deg = 180/np.pi
        
        t0 = Table((nums, nms), names=('number','name'))
        cnames=('ra','dec','x','y','z','vx','vy','vz','lightdist')
        us = (u.deg, u.deg, u.au, u.au, u.au, u.au/u.day, u.au/u.day, u.au/u.day, u.au)
        cols = [column*us[i] for i, column in enumerate(coords.T)]
        t = QTable(cols, names=cnames)
        t2= QTable([alpha*to_deg*u.deg, elong*to_deg*u.deg, Vs], names=['phase', 'elongation', 'V'])
    
        return hstack([t0, t, t2])  

    def query_2b_db(self, field_center, radius, **kwargs):
        """
        Query for asteroides visible in a field solving the 2-body equation from the asteroids'
        orbital parameters saved in the SQL database.
        
        Parameters:
                    field_center: astropy.coordinates.SkyCoord element with center coordinate of field.
                    radius: astropy.units.Quantity with radius of field.
                    
        Optional Parameters:
                    epoch: astropy.time.Time with epoch for query. Default to 'now'.
                    observer: astropy.coordinates.EarthLocation with observer's position. Default to geocenter.
                    njobs: Number of parallel processes to calculate asteroids' apparent positions. Default: 1.
                    ast_names: Asteroids' names (from Query.asts_db). Ephemerides will be calculated only for
                               those asteroids.
                    
        Returns:
                 astropy.table with computed ephemerides.
                  column 'number': Asteroid's number (from MPC).
                  column 'name': Asteroid's name (from MPC).
                  column 'ra': Asteroid's apparent right ascension (as seen by the observer).
                  column 'dec': Asteroid's apparent declination (as seen by the observer).
                  column 'x': Asteroid's heliocentric cartesian 'x' coordinate at light-emission time.
                  column 'y': Asteroid's heliocentric cartesian 'y' coordinate at light-emission time.
                  column 'z': Asteroid's heliocentric cartesian 'z' coordinate at light-emission time.
                  column 'vx': Asteroid's heliocentric cartesian 'vx' velocity at light-emission time.
                  column 'vy': Asteroid's heliocentric cartesian 'vy' velocity at light-emission time.
                  column 'vz': Asteroid's heliocentric cartesian 'vz' velocity at light-emission time.
                  column 'lightdist': Distance travel by light from asteroid's coordinate until observer.
        """
        if self.service == 'MPC': raise ValueError("'service' must be 'Lowell'.")
        # Getting all necessary parameters
        if 'observer' in kwargs: observer = kwargs['observer']
        else: observer = EarthLocation.from_geocentric(0,0,0, unit='m')
        if 'epoch' in kwargs: epoch = kwargs['epoch']
        else: epoch = Time.now()
        if 'njobs' in kwargs: njobs = kwargs['njobs']
        else: njobs = 1
        if 'ast_names' in kwargs:
            if isinstance(kwargs['ast_names'], (list,np.ndarray)): ast_names = list(kwargs['ast_names'])
            elif isinstance(kwargs['ast_names'], (str)): ast_names = [kwargs['ast_names']]
            else:
                print("ERROR in 'ast_names'"); return
            allasts = False
        else: allasts = True
        c = self.cursor; conn = self.connection
        [geopos,geovel] = observer.get_gcrs_posvel(epoch)
        obscoords = SkyCoord(geopos, frame='gcrs', obstime=epoch)
        obsx = obscoords.hcrs.cartesian.x.to('au').value
        obsy = obscoords.hcrs.cartesian.y.to('au').value
        obsz = obscoords.hcrs.cartesian.z.to('au').value
        tq = epoch.tdb.jd

        # Looking for parameters table at nearest time
        c.execute("SELECT table_title, time, ABS(time-?) FROM param_times ORDER BY ABS(time-?)", (tq,tq))
        (table, tr, terr) = c.fetchone()

        # Loading orbital parameters from database
        if allasts: qw = 'SELECT * FROM %s' %table
        else: qw = 'SELECT * FROM %s WHERE name IN (%s)' %(table,str(ast_names)[1:-1])
        tb = pd.read_sql(qw, conn, index_col='name')
        params = tb.values
        
        # Getting apparent coordinates
        args0 = tq, obsx, obsy, obsz, tr
        if njobs==1:
            coords = np.array([pycutils.apparent_coords_equatorial_heliocentric(*args0,*args1) for args1 in params])
        else:
            args = [(*args0,*args1) for args1 in params]
            pool = mp.Pool(processes=njobs)
            coords = np.array(pool.starmap(pycutils.apparent_coords_equatorial_heliocentric, args))
            pool.close()
            
        # Getting final coordinates
        ra = coords[:,0]; dec = coords[:,1]
        dists = angular_separation(field_center.ra, field_center.dec, ra*u.deg, dec*u.deg)
        infield = dists.to('rad').value<radius.to('rad').value
        coords = coords[infield]
        nms = tb.index.values[infield]; info = self.asts_db.loc[nms]
        nums = info.number.values; hs = info.H.values; gs = info.G.values
        
        [x ,y, z] = coords[:,2:5].T; geodist = coords[:,8] # au
        xe, ye, ze = x-obsx, y-obsy, z-obsz
        ast_geo = [-xe, -ye, -ze]  # Obs vector from ast
        ast_sun = [-x, -y, -z]     # Sun vector from ast
        alpha = angle_between_vectors(ast_geo, ast_sun)  # Phase angle (in rads)
        phase_func1 = np.exp(-3.33*np.tan(alpha/2)**0.36)
        phase_func2 = np.exp(-1.87*np.tan(alpha/2)**1.22)
        phase_func = (1-gs)*phase_func1 + gs*phase_func2
        Vs = hs + 5*np.log10(geodist*np.sqrt(x*x+y*y+z*z)) - 2.5*np.log10(phase_func)
        
        geo_ast = [xe, ye, ze]            # Obs vector from ast
        geo_sun = [-obsx, -obsy, -obsz]   # Sun vector from ast
        elong = angle_between_vectors(geo_ast, geo_sun)  # Elongation angle (in rads)
        to_deg = 180/np.pi
        
        t0 = Table((nums, nms), names=('number','name'))
        cnames=('ra','dec','x','y','z','vx','vy','vz','lightdist')
        us = (u.deg, u.deg, u.au, u.au, u.au, u.au/u.day, u.au/u.day, u.au/u.day, u.au)
        cols = [column*us[i] for i, column in enumerate(coords.T)]
        t = QTable(cols, names=cnames)
        t2= QTable([alpha*to_deg*u.deg, elong*to_deg*u.deg, Vs], names=['phase', 'elongation', 'V'])
        return hstack([t0, t, t2])


    def query_nb_db(self, field_center, radius, **kwargs):
        """
        Query for asteroides visible in a field. Asteoids ephemerides are calculated integrating them
        as massless particles under the influence of the Sun, the planets and Pluto. 
        Asteroids integrated are only those from nearby the requested field at the closest epoch 
        stored in the database.
    
        Parameters:
                    field_center: astropy.coordinates.SkyCoord element with center coordinate of field.
                    radius: astropy.units.Quantity with radius of field.
                    
        Optional Parameters:
                    epoch: astropy.time.Time with epoch for query. Default to 'now'.
                    observer: astropy.coordinates.EarthLocation with observer's position. Default to geocenter.
                    njobs: Number of parallel processes to calculate asteroids' apparent positions. Default: 1.
                    confidence_radius: Radius to query database. It should be greater than 'radius'. 
                                       Default: 5 degs.
                    ast_names: Asteroids' names (from Query.asts_db). Ephemerides will be calculated only for
                               those asteroids.
                    
        Returns:
                 astropy.table with computed ephemerides:
                  column 'number': Asteroid's number (from MPC).
                  column 'name': Asteroid's name (from MPC).
                  column 'ra': Asteroid's apparent right ascension (as seen by the observer).
                  column 'dec': Asteroid's apparent declination (as seen by the observer).
                  column 'x': Asteroid's heliocentric cartesian 'x' coordinate at light-emission time.
                  column 'y': Asteroid's heliocentric cartesian 'y' coordinate at light-emission time.
                  column 'z': Asteroid's heliocentric cartesian 'z' coordinate at light-emission time.
                  column 'vx': Asteroid's heliocentric cartesian 'vx' velocity at light-emission time.
                  column 'vy': Asteroid's heliocentric cartesian 'vy' velocity at light-emission time.
                  column 'vz': Asteroid's heliocentric cartesian 'vz' velocity at light-emission time.
                  column 'lightdist': Distance travel by light from asteroid's coordinate until observer.
        """
        if self.service == 'MPC': raise ValueError("'service' must be 'Lowell'.")
        # Getting all necessary parameters
        if 'observer' in kwargs: observer = kwargs['observer']
        else: observer = EarthLocation.from_geocentric(0,0,0, unit='m')
        if 'epoch' in kwargs: epoch = kwargs['epoch']
        else: epoch = Time.now()
        if 'confidence_radius' in kwargs: confidence_radius = kwargs['confidence_radius'].to('deg').value
        else: confidence_radius = radius.to('deg').value + 5
        if 'njobs' in kwargs: njobs = kwargs['njobs']
        else: njobs = 1
        if 'ast_names' in kwargs:
            if isinstance(kwargs['ast_names'], (list,np.ndarray)): ast_names = list(kwargs['ast_names'])
            elif isinstance(kwargs['ast_names'], (str)): ast_names = [kwargs['ast_names']]
            else:
                print("ERROR in 'ast_names'"); return
            allasts = False
        else: allasts = True
        c = self.cursor; conn = self.connection
        [geopos,geovel] = observer.get_gcrs_posvel(epoch)
        obscoords = SkyCoord(geopos, frame='gcrs', obstime=epoch)
        obsx = obscoords.hcrs.cartesian.x.to('au').value
        obsy = obscoords.hcrs.cartesian.y.to('au').value
        obsz = obscoords.hcrs.cartesian.z.to('au').value
        tq = epoch.tdb.jd
        
        # Looking for table at nearest time
        #c.execute("select table_title, time from times where abs(time-%f)<.25" %(tq))
        #[(table, t0)] = c.fetchall()
        c.execute("SELECT table_title, time, ABS(time-?) FROM state_times ORDER BY ABS(time-?)", (tq,tq))
        (table, t0, terr) = c.fetchone()
        if terr>0.25: 
            w = "Queried epoch is %.2f days away from closest epoch in the ephemeris database. " %terr
            w+= "User may want to increase 'db_query_radius' value."
            warnings.warn(w)
        
        # Looking for bodies near field
        ## Bodies' cartesian coordinates
        
        # Loading states from database
        if allasts: qw = 'SELECT * FROM %s' %table
        else: qw = 'SELECT * FROM %s WHERE name IN (%s)' %(table,str(ast_names)[1:-1])
        tb = pd.read_sql(qw, conn, index_col='name')
        states = tb.values; nms = tb.index.values
        
        x,y,z=states.T[:3]
        ## Bodies' equatorial coordinates from observer
        xe = x-obsx; ye = y-obsy; ze = z-obsz
        ra_geo  = np.arctan2( ye, xe ) * 180/np.pi; dec_geo = np.arctan2( ze, np.sqrt(xe*xe+ye*ye) ) * 180/np.pi
        coords_db = SkyCoord(ra_geo, dec_geo, unit='deg')
        dists = field_center.separation(coords_db)
        near = dists.deg < confidence_radius
        states = states[near]; nms = nms[near]
        
        # Calculating bodies' positions at requested epoch
        ## Solar system states
        t0_py = Time(t0, format='jd', scale='tdb')
        ss_states = get_SS_states(t0_py, frame='helio')
    
        dt = tq-t0
        args_ = ss_states, obsx, obsy, obsz, dt
        if njobs == 1:
            ephs = []
            for i, state in enumerate(states):
                ephs.append(get_apparent_coordinates(state, *args_))
            ephs = np.array(ephs)
        else:
            pool = mp.Pool(processes=njobs)
            arguments = []
            for state in states:
                arguments.append((state, *args_))
            ephs = np.array(pool.starmap(get_apparent_coordinates, arguments))
            pool.close()
        
        # Getting final coordinates
        ephs = np.array(ephs)
        ra = ephs[:,0]; dec = ephs[:,1]
        dists = angular_separation(field_center.ra.rad, field_center.dec.rad, ra, dec)
        infield = dists<radius.to('rad').value
        ephs = ephs[infield]; ephs[:,:2] = ephs[:,:2]*180/np.pi
        nms = nms[infield]; info = self.asts_db.loc[nms]
        nums = info.number.values; hs = info.H.values; gs = info.G.values
        
        [x ,y, z] = ephs[:,2:5].T; geodist = ephs[:,8] # au
        xe, ye, ze = x-obsx, y-obsy, z-obsz
        ast_geo = [-xe, -ye, -ze]  # Obs vector from ast
        ast_sun = [-x, -y, -z]     # Sun vector from ast
        alpha = angle_between_vectors(ast_geo, ast_sun)  # Phase angle (in rads)
        phase_func1 = np.exp(-3.33*np.tan(alpha/2)**0.36)
        phase_func2 = np.exp(-1.87*np.tan(alpha/2)**1.22)
        phase_func = (1-gs)*phase_func1 + gs*phase_func2
        Vs = hs + 5*np.log10(geodist*np.sqrt(x*x+y*y+z*z)) - 2.5*np.log10(phase_func)
        #return coords, nums, nms
        
        geo_ast = [xe, ye, ze]            # Obs vector from ast
        geo_sun = [-obsx, -obsy, -obsz]   # Sun vector from ast
        elong = angle_between_vectors(geo_ast, geo_sun)  # Elongation angle (in rads)
        to_deg = 180/np.pi
        
        t0 = Table((nums, nms), names=('number','name'))
        cnames=('ra','dec','x','y','z','vx','vy','vz','lightdist')
        us = (u.deg, u.deg, u.au, u.au, u.au, u.au/u.day, u.au/u.day, u.au/u.day, u.au)
        cols = [column*us[i] for i, column in enumerate(ephs.T)]
        t = QTable(cols, names=cnames)
        t2= QTable([alpha*to_deg*u.deg, elong*to_deg*u.deg, Vs], names=['phase', 'elongation', 'V'])
        return hstack([t0, t, t2])
    
    def query_mixed_db(self, field_center, radius, **kwargs):
        """
        Query for asteroides visible in a field. Asteroids ephemeris are pre-calculated using
        their orbital parameters. The ephemeris of tose asteroids that are within a 'confidence
        radius' (larger than the query's radius) are then integrated to increase their accuracy.
        Ephemeris integrations consider asteroids as massless particles under the influence of
        the Sun, the planets and Pluto.
    
        Parameters:
                    field_center: astropy.coordinates.SkyCoord element with center coordinate of field.
                    radius: astropy.units.Quantity with radius of field.
                    
        Optional Parameters:
                    epoch: astropy.time.Time with epoch for query. Default to 'now'.
                    observer: astropy.coordinates.EarthLocation with observer's position. Default to geocenter.
                    njobs: Number of parallel processes to calculate asteroids' apparent positions. Default: 1.
                    confidence_radius: The 2-Body ephemeris of the asteroids within this radius (around
                                       'field_center') are selected to be integrated. It should be greater
                                       than 'radius'. Default: 5 degs.
                    ast_names: Asteroids' names (from Query.asts_db). Ephemerides will be calculated only for
                               those asteroids.
                    
        Returns:
                 astropy.table with computed ephemerides:
                  column 'number': Asteroid's number (from MPC).
                  column 'name': Asteroid's name (from MPC).
                  column 'ra': Asteroid's apparent right ascension (as seen by the observer).
                  column 'dec': Asteroid's apparent declination (as seen by the observer).
                  column 'x': Asteroid's heliocentric cartesian 'x' coordinate at light-emission time.
                  column 'y': Asteroid's heliocentric cartesian 'y' coordinate at light-emission time.
                  column 'z': Asteroid's heliocentric cartesian 'z' coordinate at light-emission time.
                  column 'vx': Asteroid's heliocentric cartesian 'vx' velocity at light-emission time.
                  column 'vy': Asteroid's heliocentric cartesian 'vy' velocity at light-emission time.
                  column 'vz': Asteroid's heliocentric cartesian 'vz' velocity at light-emission time.
                  column 'lightdist': Distance travel by light from asteroid's coordinate until observer.
        """
        if self.service == 'MPC': raise ValueError("'service' must be 'Lowell'.")
        # Getting all necessary parameters
        if 'observer' in kwargs: observer = kwargs['observer']
        else: observer = EarthLocation.from_geocentric(0,0,0, unit='m')
        if 'epoch' in kwargs: epoch = kwargs['epoch']
        else: epoch = Time.now()
        if 'njobs' in kwargs: njobs = kwargs['njobs']
        else: njobs = 1
        if 'confidence_radius' in kwargs: confidence_radius = kwargs['confidence_radius'].to('deg').value
        else: confidence_radius = radius.to('deg').value + 3
        if 'ast_names' in kwargs:
            if isinstance(kwargs['ast_names'], (list,np.ndarray)): ast_names = list(kwargs['ast_names'])
            elif isinstance(kwargs['ast_names'], (str)): ast_names = [kwargs['ast_names']]
            else:
                print("ERROR in 'ast_names'"); return
            allasts = False
        else: allasts = True
        c = self.cursor; conn = self.connection
        [geopos,geovel] = observer.get_gcrs_posvel(epoch)
        obscoords = SkyCoord(geopos, frame='gcrs', obstime=epoch)
        obsx = obscoords.hcrs.cartesian.x.to('au').value
        obsy = obscoords.hcrs.cartesian.y.to('au').value
        obsz = obscoords.hcrs.cartesian.z.to('au').value
        tq = epoch.tdb.jd
        
        # Looking for orbital parameters table and states table at nearest time
        c.execute("SELECT table_title, time, ABS(time-?) FROM param_times ORDER BY ABS(time-?)", (tq,tq))
        (table_params, tr, terr) = c.fetchone()
        table_states = 'state_'+table_params.split('_')[1]
        
        # Loading orbital parameters from database
        if allasts: qw = "SELECT * FROM %s" %table_params
        else: qw = "SELECT * FROM %s WHERE name IN (%s)" %(table_params,str(ast_names)[1:-1])
        tb_params = pd.read_sql(qw, conn, index_col='name')
        params = tb_params.values; nms = tb_params.index.values
        
        # Getting apparent coordinates using 2-body solution
        #args0 = tq, obsx, obsy, obsz, tr
        #coords = np.array([pycutils.apparent_coords_equatorial_heliocentric(*args0,*args1) for args1 in params])
        #ra = coords[:,0]; dec = coords[:,1]
        
        # Getting instantaneous states using 2-body solution (about 3s faster than getting apparent coordinates)
        args0 = tq, tr
        states_ = np.array([pycutils.state_equatorial_heliocentric(*args0,*args1) for args1 in params])
        [xh ,yh, zh] = np.array(states_)[:,:3].T
        # Instantaneous ephemerides are calculated for all asteroids.
        xe = xh-obsx; ye = yh-obsy; ze = zh-obsz
        ra  = np.arctan2( ye, xe ); dec = np.arctan2( ze, np.sqrt(xe*xe+ye*ye) )
        
        # Selecting bodies in confidence radius for n-body integration
        dists = angular_separation(field_center.ra.rad, field_center.dec.rad, ra, dec)
        inconffield = dists<(confidence_radius*np.pi/180)
        nms = nms[inconffield]
        
        # Loading states from database
        if allasts: qw = 'SELECT * FROM %s' %table_states
        else: qw = 'SELECT * FROM %s WHERE name IN (%s)' %(table_states,str(list(nms))[1:-1])
        tb_states = pd.read_sql(qw, conn, index_col='name').loc[nms]
        states = tb_states.values
        
        # Calculating bodies' positions at requested epoch
        ## Solar system states
        tr_py = Time(tr, format='jd', scale='tdb')
        ss_states = get_SS_states(tr_py, frame='helio')
        
        dt = tq-tr
        args0 = ss_states, obsx, obsy, obsz, dt
        if njobs == 1:
            ephs = np.array([get_apparent_coordinates(state, *args0) for state in states])
        else:
            pool = mp.Pool(processes=njobs)
            arguments = [(state, *args0) for state in states]
            ephs = np.array(pool.starmap(get_apparent_coordinates, arguments))
            pool.close()
        
        # Getting final coordinates
        ra = ephs[:,0]; dec = ephs[:,1]
        dists = angular_separation(field_center.ra.rad, field_center.dec.rad, ra, dec)
        infield = dists<radius.to('rad').value
        ephs = ephs[infield]; ephs[:,:2] = ephs[:,:2]*180/np.pi
        nms = nms[infield]; info = self.asts_db.loc[nms]
        nums = info.number.values; hs = info.H.values; gs = info.G.values
        
        [x, y, z] = ephs[:,2:5].T; geodist = ephs[:,8] # au
        xe, ye, ze = x-obsx, y-obsy, z-obsz
        ast_geo = [-xe, -ye, -ze]  # Obs vector from ast
        ast_sun = [-x, -y, -z]     # Sun vector from ast
        alpha = angle_between_vectors(ast_geo, ast_sun)  # Phase angle (in rads)
        phase_func1 = np.exp(-3.33*np.tan(alpha/2)**0.36)
        phase_func2 = np.exp(-1.87*np.tan(alpha/2)**1.22)
        phase_func = (1-gs)*phase_func1 + gs*phase_func2
        Vs = hs + 5*np.log10(geodist*np.sqrt(x*x+y*y+z*z)) - 2.5*np.log10(phase_func)
        #return coords, nums, nms
        
        geo_ast = [xe, ye, ze]            # Obs vector from ast
        geo_sun = [-obsx, -obsy, -obsz]   # Sun vector from ast
        elong = angle_between_vectors(geo_ast, geo_sun)  # Elongation angle (in rads)
        to_deg = 180/np.pi
        
        t0 = Table((nums, nms), names=('number','name'))
        cnames=('ra','dec','x','y','z','vx','vy','vz','lightdist')
        us = (u.deg, u.deg, u.au, u.au, u.au, u.au/u.day, u.au/u.day, u.au/u.day, u.au)
        cols = [column*us[i] for i, column in enumerate(ephs.T)]
        t = QTable(cols, names=cnames)
        t2= QTable([alpha*to_deg*u.deg, elong*to_deg*u.deg, Vs], names=['phase', 'elongation', 'V'])
        return hstack([t0, t, t2]) 
    
    def close(self): self.connection.close()
        
    def asteph(self, ast, **kwargs):
        """
        Returns ephemerides from specified asteroid at given epochs.
        Ephemerides are calculated performing an n-body simulation that considers
        the sun, all 8 planets and the asteroid as a massless particle, starting from
        a time where all states are known and integrating all bodies position until
        requested time (correcting by light aberration).
    
        Parameters:
                    ast: Name of the asteroid (str) or it index number (int). If a string is given,
                         a perfect match of the name is looked for at the Lowell's database. If a number
                         is given, it is considered as the index (starting from zero) of the asteroid's
                         in the Lowell' database (so, for numbered bodies, if you want '(1) Ceres',
                         the index is zero).
                    
        Optional Parameters:
                    epochs: astropy.time.Time with epoch(s) to calculate ephemerides. Default to 'now'.
                    observer: astropy.coordinates.EarthLocation with observer's position. Default to geocenter.
                    
        Returns:
                 astropy.table with computed ephemerides:
                  column 'epoch_jd': Epoch in julian days (in UTC).
                  column 'ra': Asteroid's apparent right ascension (as seen by the observer).
                  column 'dec': Asteroid's apparent declination (as seen by the observer).
                  column 'x': Asteroid's heliocentric cartesian 'x' coordinate at light-emission time.
                  column 'y': Asteroid's heliocentric cartesian 'y' coordinate at light-emission time.
                  column 'z': Asteroid's heliocentric cartesian 'z' coordinate at light-emission time.
                  column 'vx': Asteroid's heliocentric cartesian 'vx' velocity at light-emission time.
                  column 'vy': Asteroid's heliocentric cartesian 'vy' velocity at light-emission time.
                  column 'vz': Asteroid's heliocentric cartesian 'vz' velocity at light-emission time.
                  column 'lightdist': Distance travel by light from asteroid's coordinate until observer.
                  column 'phase': Angle between the Sun and the Observer from the asteroid.
                  column 'elongation': Angle between the Sun and the asteroid from the Obsever.
                  column 'V':  Apparent magnitude as seen by the observer in V band (calculated from the
                               asteroid's absolute magnitude in Lowell's database).
        """
        if self.service == 'MPC': raise ValueError("'service' must be 'Lowell'.")
        # Getting all necessary parameters
        if isinstance(ast, (int,np.int64)): ast_name = self.asts_db.iloc[ast].name
        if isinstance(ast, str): ast_name = ast
        
        if 'observer' in kwargs: observer = kwargs['observer']
        else: observer = EarthLocation.from_geocentric(0,0,0, unit='m')
        
        if 'epochs' in kwargs:
            if not isinstance(kwargs['epochs'], Time):
                print('ERROR'); return
            if isinstance(kwargs['epochs'].jd, np.ndarray):
                epochs0 = kwargs['epochs']
            else: epochs0 = Time([kwargs['epochs']])
        else: epochs0 = Time([Time.now().jd], format='jd')
        
        c = self.cursor
        iepochs = np.argsort(epochs0.jd); epochs = epochs0[iepochs] # Epochs are sorted to speed up integration
        tqs  = epochs.tdb.jd
        tmin = min(tqs); tmax = min(tqs)
        
        # Looking for table at nearest time
        c.execute("SELECT table_title, time, ABS(time-?) FROM state_times ORDER BY ABS(time-?)", (tmin,tmin))
        (table_min, t0_min, terr_min) = c.fetchone()
        c.execute("SELECT table_title, time, ABS(time-?) FROM state_times ORDER BY ABS(time-?)", (tmax,tmax))
        (table_max, t0_max, terr_max) = c.fetchone()
        if terr_min < terr_max:
            t0 = t0_min; table = table_min; dir_integ = 1
        else:
            t0 = t0_max; table = table_max; dir_integ = -1; tqs = tqs[::-1]; epochs = epochs[::-1]; iepochs = iepochs[::-1]
        
        # Observatory coordinates
        [geopos,geovel] = observer.get_gcrs_posvel(epochs)
        obscoords = SkyCoord(geopos, frame='gcrs', obstime=epochs)
        obsx = obscoords.hcrs.cartesian.x.to('au').value
        obsy = obscoords.hcrs.cartesian.y.to('au').value
        obsz = obscoords.hcrs.cartesian.z.to('au').value
        
        # Asteroid's state at nearest time
        c.execute('SELECT x,y,z,vx,vy,vz FROM %s WHERE name="%s"' %(table,ast_name))
        state_ast = c.fetchone()
        
        # Calculating bodies' positions at requested epoch
        ## Solar system states
        t0_py = Time(t0, format='jd', scale='tdb')
        ss_states = get_SS_states(t0_py, frame='helio')
            
        # ASTEROID'S INTEGRATION
        ## Creating simulation
        sim = get_SS_sim(ss_states); sim.N_active = sim.N # Solar system
        sim.add(x=state_ast[0], y=state_ast[1], z=state_ast[2], vx=state_ast[3], vy=state_ast[4], vz=state_ast[5]) # Asteroid
        sun = sim.particles[0]; ast = sim.particles[-1]; sim.dt = 0.1*dir_integ
        
        #return tqs, t0
        delta_tdbs = tqs - t0
        xhs, yhs, zhs, vxhs, vyhs, vzhs, lightdist = [], [], [], [], [], [], []
        for i, delta_tdb in enumerate(delta_tdbs):
            # Integration until observation time
            sim.integrate(delta_tdb)
            # Correcting by light's travel time
            gdist = np.sqrt((ast.x-sun.x-obsx[i])**2+(ast.y-sun.y-obsy[i])**2+(ast.z-sun.z-obsz[i])**2)
            lighttime = gdist*au_to_d
            t3 = delta_tdb - lighttime
            niter=0
            while niter<31:
                niter += 1; t2=t3
                sim.integrate(t2)
                gdist = np.sqrt((ast.x-sun.x-obsx[i])**2+(ast.y-sun.y-obsy[i])**2+(ast.z-sun.z-obsz[i])**2)
                lighttime = gdist*au_to_d
                t3 = delta_tdb - lighttime
                if abs (t3 - t2) < 1.0e-10: break
            xhs.append(ast.x-sun.x)   ; yhs.append(ast.y-sun.y)   ; zhs.append(ast.z-sun.z)
            vxhs.append(ast.vx-sun.vx); vyhs.append(ast.vy-sun.vy); vzhs.append(ast.vz-sun.vz)
            lightdist.append(lighttime/au_to_d)  # In au
            
        # Ordering data to original epochs' order
        N = len(epochs)
        obx,oby,obz,x,y,z = np.zeros(N),np.zeros(N),np.zeros(N),np.zeros(N),np.zeros(N),np.zeros(N)
        vx,vy,vz,geodist = np.zeros(N),np.zeros(N),np.zeros(N),np.zeros(N)
        obx[iepochs]=obsx; oby[iepochs]=obsy; obz[iepochs]=obsz; x[iepochs]=xhs; y[iepochs]=yhs; z[iepochs]=zhs
        vx[iepochs]=vxhs; vy[iepochs]=vyhs; vz[iepochs]=vzhs; geodist[iepochs]=lightdist
        
        # Getting final data
        c.execute('SELECT number,H,G FROM asts_info WHERE name="%s"' %(ast_name))
        num,h,g = c.fetchone()
        
        xe = x-obx; ye = y-oby; ze = z-obz
        ra  = np.arctan2( ye, xe ); dec = np.arctan2( ze, np.sqrt(xe*xe+ye*ye) ) # In rads
        neg_ra = ra < 0; ra[neg_ra] += pycutils.TWOPI # In rads
        to_deg = 180/np.pi; ra *= to_deg; dec *= to_deg
        
        ast_geo = [-xe, -ye, -ze]  # Obs vector from ast
        ast_sun = [-x, -y, -z]     # Sun vector from ast
        alpha = angle_between_vectors(ast_geo, ast_sun)  # Phase angle (in rads)
        phase_func1 = np.exp(-3.33*np.tan(alpha/2)**0.36)
        phase_func2 = np.exp(-1.87*np.tan(alpha/2)**1.22)
        phase_func = (1-g)*phase_func1 + g*phase_func2
        Vs = h + 5*np.log10(geodist*np.sqrt(x*x+y*y+z*z)) - 2.5*np.log10(phase_func)
        
        geo_ast = [xe, ye, ze]         # Obs vector from ast
        geo_sun = [-obx, -oby, -obz]   # Sun vector from ast
        elong = angle_between_vectors(geo_ast, geo_sun)*to_deg  # Elongation angle (in degs)
        
        #t0 = Table((nums, nms), names=('number','name'))
        deg, au, au_d = u.deg, u.au, u.au/u.day
        cnames=('epoch_jd', 'ra','dec','x','y','z','vx','vy','vz','lightdist','phase','elongation', 'V')
        cols = [epochs0.jd, ra*deg, dec*deg, x*au, y*au, z*au, vx*au_d, vy*au_d, vz*au_d, geodist*au,
                alpha*to_deg*deg, elong*deg, Vs]
        t = QTable(cols, names=cnames)
        
        return t



import astropy.constants as const
from astropy.coordinates import SkyCoord, get_body_barycentric, get_body_barycentric_posvel, solar_system_ephemeris, EarthLocation, GCRS
from astropy import units as u
from . import pycutils
import rebound
import numpy as np

solar_system_ephemeris.set('de432s')

au_to_d = (1*u.au/const.c.to('au/s')).to('d').value
def get_apparent_coordinates(state_ast, states_ss, obsx, obsy, obsz, delta_tdb):
    # Creating simulation
    sim = get_SS_sim(states_ss); sim.N_active = sim.N # Solar system
    sim.add(x=state_ast[0], y=state_ast[1], z=state_ast[2], vx=state_ast[3], vy=state_ast[4], vz=state_ast[5]) # Asteroid
    sun = sim.particles[0]; ast = sim.particles[-1]; #sim.dt = -0.1
    # Integration until observation time
    sim.integrate(delta_tdb)
    # Correcting by light's travel time
    gdist = np.sqrt((ast.x-sun.x-obsx)**2+(ast.y-sun.y-obsy)**2+(ast.z-sun.z-obsz)**2)
    lighttime = gdist*au_to_d
    t3 = delta_tdb - lighttime
    niter=0
    while niter<31:
        niter += 1; t2=t3
        sim.integrate(t2)
        gdist = np.sqrt((ast.x-sun.x-obsx)**2+(ast.y-sun.y-obsy)**2+(ast.z-sun.z-obsz)**2)
        lighttime = gdist*au_to_d
        t3 = delta_tdb - lighttime
        if abs (t3 - t2) < 1.0e-10: break
    xh , yh, zh =  ast.x-sun.x , ast.y-sun.y , ast.z-sun.z
    vxh,vyh,vzh = ast.vx-sun.vx,ast.vy-sun.vy,ast.vz-sun.vz
    xe = xh-obsx; ye = yh-obsy; ze = zh-obsz
    ra_geo  = np.arctan2( ye, xe ); dec_geo = np.arctan2( ze, np.sqrt(xe*xe+ye*ye) )
    if ra_geo < 0: ra_geo += pycutils.TWOPI
    return ra_geo, dec_geo, xh,yh,zh, vxh,vyh,vzh, lighttime/au_to_d

################################################################

def get_SS_states(epoch, planets=True, frame='bary'):
    """
    Computes states (positions and velocities) for the Solar System in a given epoch.
    Parameters:
                epoch: astropy.time.Time value.
                planets: Bool. If True, returns states of all planets (and Pluto).
                         If False, returns only Sun's state. Default True.
                frame: Which coordinate frame to use: 'bary' for barycentric and
                       'helio' for heliocentric. Default 'bary'.
    Returns: numpy.array with the solar system states
    """
    names = ['mercury', 'venus', 'earth-moon-barycenter', 'mars', 'jupiter', 'saturn', 'uranus', 'neptune']#, 'pluto']
    states = []
    # Sun
    sun_pos_bary, sun_vel_bary = get_body_barycentric_posvel('sun', epoch)
    x = sun_pos_bary.x.to('au').value; y = sun_pos_bary.y.to('au').value; z = sun_pos_bary.z.to('au').value
    vx = sun_vel_bary.x.to('au/d').value; vy = sun_vel_bary.y.to('au/d').value; vz = sun_vel_bary.z.to('au/d').value
    if frame == 'bary': states.append([x,y,z,vy,vy,vz])
    elif frame == 'helio': states.append([0,0,0,0,0,0])
    else: raise ValueError("'frame' must be 'bary' or 'helio'")
    # Planets
    if planets:
        for i, name in enumerate(names):
            pos_bary, vel_bary = get_body_barycentric_posvel(name, epoch)
            if frame=='bary':
                pos = pos_bary; vel = vel_bary
            else:
                pos = pos_bary-sun_pos_bary; vel = vel_bary-sun_vel_bary
            x = pos.x.to('au').value; y = pos.y.to('au').value; z = pos.z.to('au').value
            vx = vel.x.to('au/d').value; vy = vel.y.to('au/d').value; vz = vel.z.to('au/d').value
            states.append([x,y,z,vx,vy,vz])
    return np.array(states)
def get_SS_sim(ss_states, planets=True):
    """
    Starts a Solar System simulation for integration.
    Parameters: 
                ss_states: numpy.array with solar system states
                           tu start simulation.
                planets: Bool. If False, considers only the Sun.
                         Default True.
    Returns: rebound.Simulation
    """
    masses = np.array([1, 1.6601141530543488e-07, 2.4478382877847715e-06, 3.040432648022642e-06, 3.2271560375549977e-07,
              0.0009547919152112404, 0.0002858856727222417, 4.36624373583127e-05, 5.151383772628674e-05])#, 7.361781606089469e-09])
    if not planets: masses = masses[:1]
    sim = rebound.Simulation()
    sim.units = ('d', 'AU', 'Msun')
    for i in range(len(masses)):
        st = ss_states[i]
        sim.add(m=masses[i],x=st[0], y=st[1], z=st[2], vx=st[3], vy=st[4], vz=st[5])
    return sim

def get_integration(states, ss_states, dt, itimes):
    """
    From initial states of asteroids at epoch0, returns dictionary with integrated
    positions of those asteroids at specified relative times from epoch0.
    Integration is performed using rebound.
    
    Parameters:
                states: List of first equatorial heliocentric states of all bodies to integrate from.
                ss_states: List of first equatorial heliocentric states of all 8 planets.
                dt: Maximum time-step for integration.
                relative_times: List with times (in days) relative to first states' time to integrate.
    Returns: dicctionary with states, with the relative epoch (tdb.jd) as key.
    """
    # Starting Solar System
    sim = get_SS_sim(ss_states)
    sim.N_active = sim.N
    # Adding asteroids
    for state in states:
        xh, yh, zh, vxh, vyh, vzh = state
        sim.add(x=xh, y=yh, z=zh, vx=vxh, vy=vyh, vz=vzh)
    # Integration for each time
    sim.dt = dt; sun = sim.particles[0]
    dicc_states = {}
    for t in itimes:
        sim.integrate(t); sim.move_to_hel()
        statesi = np.zeros((sim.N,6),dtype="float64")
        sim.serialize_particle_data(xyzvxvyvz=statesi)
        dicc_states[t] = statesi[9:] #[10:] #[9:] without Pluto; [10:] with Pluto
    return dicc_states
################################################################
def angle_between_vectors(x1, x2):
    """
    Computes angle between vectors x1 and x2 using the dot and the cross product.
    
    Parameters:
                x1: array_like with _x_, _y_ and _z_ coordinates.
                x2: array_like with _x_, _y_ and _z_ coordinates.
                
    Returns:
             array with angle between x1 and x2 in radians.
    """
    #x1mod = np.sqrt((x1[0]*x1[0]) + (x1[1]*x1[1]) + (x1[2]*x1[2]))
    #x2mod = np.sqrt((x2[0]*x2[0]) + (x2[1]*x2[1]) + (x2[2]*x2[2]))
    x1_dot_x2 = (x1[0]*x2[0]) + (x1[1]*x2[1]) + (x1[2]*x2[2])
    x1_cross_x2_i = x1[1]*x2[2] - x1[2]*x2[1]
    x1_cross_x2_j = x1[2]*x2[0] - x1[0]*x2[2]
    x1_cross_x2_k = x1[0]*x2[1] - x1[1]*x2[0]
    x1_croos_x2_mod = np.sqrt(x1_cross_x2_i*x1_cross_x2_i+x1_cross_x2_j*x1_cross_x2_j+x1_cross_x2_k*x1_cross_x2_k)
    #return abs( np.arccos(x1x2/(x1mod*x2mod)) )
    return abs( np.arctan2( x1_croos_x2_mod , x1_dot_x2 ) )
    #return abs( np.arctan( x1_croos_x2_mod / x1_dot_x2) )



import urllib, gzip
import os, time
import numpy as np
from . import pycutils
from astropy.time import Time
from astropy.coordinates import SkyCoord, get_body_barycentric, get_body_barycentric_posvel, solar_system_ephemeris, EarthLocation, GCRS
import multiprocessing as mp
import sqlite3, warnings
from .aleph_utils import *
import pandas as pd

solar_system_ephemeris.set('de432s') 

# https://stjarnhimlen.se/comp/ppcomp.html
# https://ssd.jpl.nasa.gov/txt/aprx_pos_planets.pdf

## FUNCTIONS TO READ ORBITAL PARAMETERS CATALOGUES
def unpack_number(packed_num):
    S = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz'
    try: return int(packed_num)
    except:
        if '~' in packed_num:
            pw = np.array([62**3, 62**2, 62, 1])
            num = 620000
            for i, s in enumerate(packed_num[1:]): num += S.find(s)*pw[i]
        else:
            num = str(S.find(packed_num[0])) + packed_num[1:]
        return int(num)
def getfl(s):
    try: return float(s)
    except: return np.nan
def get_astdata(s, source='MPCORB'):
    if source=='MPCORB':
        pack_num = s[:7]; H = s[8:13]; G = s[14:19]; epoch = s[20:25]; M = s[26:35]; peri = s[37:46]; node = s[48:57]
        incl = s[59:68]; e = s[70:79]; n = s[80:91]; a = s[92:103]; U = s[105:106]; ref = s[107:116]; nobs = s[117:122]
        nops = s[123:126]; arc = s[127:136]; rms = s[137:141]; perts1 = s[142:145]; perts2 = s[146:149]; compnm = s[150:160]
        flags = s[161:165]; readdes = s[166:194]; lastobs = s[194:202]
        div = readdes.split(')'); name = div[-1].strip()
        if len(div)==2: num = div[0].strip()[1:]
        else: num = '-'
        dat = {'num':num, 'H':getfl(H), 'G':getfl(G), 'epoch':unpack_epoch(epoch,source), 'M_epoch':float(M),
               'Arg_Peri':float(peri), 'Node':float(node), 'incl':float(incl), 'e':float(e), 'n':float(n), 'a':float(a),
               'U':U, 'ref':ref, 'Nobs':getfl(nobs), 'Nops':float(nops), 'arc':arc, 'rms':getfl(rms), 'perts1':perts1,
               'perts2':perts2, 'compnm':compnm, 'flags':flags, 'name':name, 'lastobs':lastobs}
    if source=='astorb':
        num = s[:6]; name = s[7:25]; compnm = s[26:41]; H = s[42:47]; G = s[48:53]; color_B_V = s[54:58]
        iras_diam_km = s[59:64]; iras_taxonomy = s[65:69]; code6 = s[70:94]; arc_days = s[95:100]; nobs = s[101:105]
        epoch = s[106:114]; M = s[115:125]; peri = s[126:136]; node = s[137:147]; incl = s[147:157]; e = s[158:168]
        a = float(s[168:181]); compdate = s[182:190]; ceu_arcsec = s[191:198]; ceurate_arcsecday = s[199:207]
        ceudate = s[208:216]; peu_arcsec = s[217:224]; peudate = s[225:233]; peu_gratest_ceunext10yr = s[234:241]
        peudate_gratest_ceunext10yr = s[242:250]; peu_gratest_peunext10yr = s[251:258]; peudate_gratest_peunext10yr = s[259:267]
        dat = {'num':num, 'name':name, 'compnm':compnm, 'H':getfl(H), 'G':getfl(G), 'color_B_V':getfl(color_B_V),
               'iras_diam_km':getfl(iras_diam_km), 'iras_taxonomy':iras_taxonomy, 'code6':code6, 'arc_days':getfl(arc_days),
               'Nobs':getfl(nobs), 'epoch':unpack_epoch(epoch,source), 'M_epoch':float(M), 'Arg_Peri':float(peri),
               'Node':float(node), 'incl':float(incl), 'e':float(e), 'a':a, 'compdate':compdate, 'ceu_arcsec':ceu_arcsec,
               'ceurate_arcsecday':ceurate_arcsecday, 'ceudate':ceudate, 'peu_gratest_ceunext10yr':peu_gratest_ceunext10yr,
               'peudate_gratest_ceunext10yr':peudate_gratest_ceunext10yr, 'peu_gratest_peunext10yr':peu_gratest_peunext10yr,
               'peudate_gratest_peunext10yr':peudate_gratest_peunext10yr}
        P = 365.2568984 * a**1.5; dat['n'] = 360.0/P
    return dat
def unpack_epoch(packed_epoch, source='MPCORB'):
    if source=='MPCORB':
        S = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ'
        y = int(str(S.find(packed_epoch[0]))+packed_epoch[1:3])
        m = S.find(packed_epoch[3]); d = S.find(packed_epoch[4])
        return '%04i-%02i-%02i' %(y, m, d)
    if source=='astorb':
        y = packed_epoch[:4]; m = packed_epoch[4:6]; d = packed_epoch[6:]
        return '%s-%s-%s' %(y, m, d)

def download_orbparams_catalogue(service, filename):
    """
    Donwloads the orbital parameter catalogue from the specified service.

    Parameters:
                service: Specifies origin of orbital parameters.
                         Can be 'Lowell' or 'MPC'.
                filename: Name of the saved file.
    """
    if service == 'MPC':
        #url = "https://www.minorplanetcenter.org/iau/MPCORB/MPCORB.DAT"
        url = 'https://minorplanetcenter.net/iau/MPCORB/MPCORB.DAT.gz'
    elif service == 'Lowell':
        #url = 'ftp://ftp.lowell.edu/pub/elgb/astorb.dat.gz'
        ve = "Downloading 'astorb.dat' via FTP is not longer supported. "
        ve+= "Download it yourself in https://asteroid.lowell.edu/main/astorb/ and"
        ve+= "provide it via 'filename' parameter."
        raise ValueError(ve)
    else: raise ValueError("'service' must be 'Lowell' or 'MPC'")

    response = urllib.request.urlopen(url)
    fo = open(filename,'wb'); fo.write(gzip.decompress(response.read())); fo.close()


PACKAGE_PATH = os.path.abspath(os.path.dirname(__file__))

class OrbParams:
    def __init__ (self, service='MPC', update=False, **kwargs):
        """
        Reads an orbital parameter file for asteroids and computes ephemerides
        
        Parameters:
                    service: Sets origin of orbital parameters. Can be 'Lowell' or 'MPC'.
                             Default to 'Lowell'.
                    update: bool. If True, it downloads the orbital parameter's file
                            from the appropiate servive and saves it in the package's
                            directory. Default to False.
        Other Parameters:
                    filename: Str. Specifies file (and path) with orbital parameters to
                              read. This file must be in Lowell's or MPC's format.
        Attributes:
                    service: 'Lowell' or 'MPC'
                    asts: List of dictionaries containing all asteroids' information.
                    dicc_epochs_M: Dictionary with reference epochs of orbital parameters.
        """
        self.service = service
        if 'filename' in kwargs: self.filename = kwargs['filename']
        else:
            if service=='Lowell': self.filename = PACKAGE_PATH + '/astorb.dat'
            if service=='MPC': self.filename = PACKAGE_PATH + '/MPCORB.DAT'
        
        if update: self.update()
        
        self.asts = self.read(self.filename)
        
        # Getting dicc of epochs
        epochs_M = []
        for ast in self.asts: epochs_M.append(ast['epoch'])
        set_epoch_M = list(set(epochs_M)); epochs_M = Time(set_epoch_M, format='iso', scale='tt')
        self.dicc_epochs_M = {t:Time(t, format='iso', scale='tt') for t in set_epoch_M}
        
    def update(self):
        """
        It downloads the orbital parameter's file from the appropiate servive.
        """
        print('UPDATING',self.filename)
        download_orbparams_catalogue(self.service, self.filename)
        print("UPDATE FINISHED!")
                
    def read(self, filename):
        """
        Reads orbital parameter file and saves its data as a dicctionary in a list.
        """
        if not os.path.exists(filename):
            err = "Orbital parameter file not found. You need to run "
            err+= "'DataBase.download_orbparams_catalogue(service, filename)' or initialize OrbParams with "
            err+= "'update=True' or initialize OrbParams setting 'filename' with an orbital parameter "
            err+= "file in a different directory."
            raise ValueError(err)
        if self.service == 'MPC':
            asts = []
            is_header=True; is_numbered=False; is_unnumbered=False; is_1oposition=False
            i = 0; fi = open(filename,'r')
            for line in fi:
                i+=1
                if is_header: pass
                if '------' in line:
                    is_numbered=True; is_header=False
                    continue
                if is_numbered and len(line)==1:
                    is_numbered=False; is_unnumbered=True
                    continue
                if is_unnumbered and len(line)==1:
                    is_unnumbered=False; is_1oposition=True
                    continue
                if is_numbered or is_unnumbered or is_1oposition:
                    asts.append(get_astdata(line, source='MPCORB'))
            fi.close()
                
        elif self.service == 'Lowell':
            fi = open(filename,'r')
            asts = []
            i = 0
            for line in fi: asts.append(get_astdata(line, source='astorb'))
            fi.close()
                    
        else: raise ValueError("'service' must be 'Lowell' or 'MPC'")
        
        return asts
    
    def get_epoch_state(self, epoch_tdb_jd, ast_indexs=None, njobs=1):
        """
        Returns states of asteroids at specified epoch in heliocentric equatorial coordinates.
        One state consists of a list containing positions and velocities: [x,y,z,vx,vy,z].
        
        Parameters:
                    epoch_tdb_jd: float Epoch. Barycentric Dynamical Time in Julian Day format.
                    ast_indexs: Index list of asteroids in AlerceEphs.asts to use.
                                If not specified, states for all asteroids are calculated.
                    njobs: int Number of parallel processes to run calculations. Default 1.
        """
        if not isinstance(ast_indexs, (list,np.ndarray)): ast_indexs = np.arange(len(self.asts))
        t = epoch_tdb_jd # calculated only once
    
        if njobs==1:
            results = []
            for idx in ast_indexs:
                ast = self.asts[idx]
                r = pycutils.state_equatorial_heliocentric(t, t, ast['a'],ast['e'],ast['incl'],ast['Node'],ast['Arg_Peri'],ast['M_epoch'])
                results.append(r)
        else:
            pool = mp.Pool(processes=njobs)
            arguments = []
            for idx in ast_indexs:
                ast = self.asts[idx]
                arguments.append((t, t, ast['a'],ast['e'],ast['incl'],ast['Node'],ast['Arg_Peri'],ast['M_epoch']))
            results = pool.starmap(pycutils.state_equatorial_heliocentric, arguments)
            pool.close()
        return results
    
    def create_states_database(self, sqlfile=PACKAGE_PATH+'/aleph_states.db', ast_indexs=None, verbose=True): #, njobs=1
        """
        Computes states of all Lowell's asteroids in an interval of 101 days around ASTORB
        reference epoch with a time interval of 0.5 days and saves them in a SQL database.
        Since orbital parameters in MPCORB.DAT can be at different epochs, this utility only
        works with astorb.dat data.
        
        Parameters:
                    sqlfile: File name of SQL database. Default 'aleph_states.db'.
                    ast_indexs: Index list of asteroids in AlerceEphs.asts to use.
                                If not specified, states for all asteroids are calculated.
                    verbose: bool If True, print messages indicating progress. Default True.

        The SQL database contains the following tables:

         * 'asts_info'   : Contains asteroids' information. Columns: 'name', 'number', 'H', 'G'.
         * 'param_T0'    : Contains the orbital parameters of all asteroids as given in 
                           astorb.dat in columns 'name', 'a', 'e', 'i', 'longnode',
                           'argperi' and 'meananom' (heliocentric ecliptic coordinates).
         * 'param_TXXXX' : Same as 'param_T0', but with values in other epochs.
                           Only available after using function 'adding_params_into_db'.
         * 'param_times' : Contains columns 'table_title' (names with tables of states) and
                           'time' with the corresponding tdb time for those states.
         * 'state_T0'    : Contains the states of every asteroid in the astorb's epoch of 
                           osculation (common for all asteroids) in the columns 'name', 'x',
                           'y', 'z', 'vx', 'vy' and 'vz' (heliocentric equatorial coordinates).
         * 'state_TXXXX' : Same as 'state_T0', but with values in other epochs.
                           Only available after using function 'adding_states_into_db'.
         * 'state_times' : Contains columns 'table_title' (names with tables of states) and
                           'time' with the corresponding tdb time for those states.
        """
#                    njobs: int Number of parallel processes to run calculations. Default 1.
        if not isinstance(ast_indexs, (list,np.ndarray)): ast_indexs = np.arange(len(self.asts))
        epoch = Time(self.asts[0]['epoch'], format='iso', scale='tt')
        #t = np.round(epoch.tdb.jd,decimals=2) # calculated only once
        t0 = epoch.tdb.jd
        
        # Creating SQL DB
        #if verbose: print('* Creating SQL Database')
        pkey = [[self.asts[i]['name'].strip()] for i in ast_indexs]
        info = [[self.asts[i]['num'], self.asts[i]['H'], self.asts[i]['G']] for i in ast_indexs]
        
        ## Starting database
        os.system('rm %s' %sqlfile) # Database is constructed again
        conn = sqlite3.connect(sqlfile)
        ## Create param_times table
        conn.execute("CREATE TABLE param_times(table_title, time)")
        ## Create state_times table
        conn.execute("CREATE TABLE state_times(table_title, time)")
        ## Create and fill asts_info table
        conn.execute("CREATE TABLE asts_info(name TEXT, number TEXT, H REAL, G REAL, PRIMARY KEY(name))")
        conn.executemany("INSERT INTO asts_info (name, number, H, G) VALUES (?,?,?,?)", np.hstack((pkey, info)) )

        # Orbital parameters at reference epoch
        if verbose: print('* Saving orbital parameters at reference epoch')
        ti = time.time()
        params = []
        for idx in ast_indexs:
            ast = self.asts[idx]
            params.append([ast['name'].strip(),ast['a'],ast['e'],ast['incl'],ast['Node'],ast['Arg_Peri'],ast['M_epoch']])
        #params = np.array(params, dtype=str)
        ### Create state table
        table_name = "param_T0"
        conn.execute("CREATE TABLE param_T0(name TEXT,a REAL,e REAL,i REAL,longnode REAL,argperi REAL,meananom REAL, PRIMARY KEY(name))")
        ### Fill times table
        conn.execute("INSERT INTO param_times(table_title, time) VALUES (?, ?)", (table_name,t0))
        ### Fill state table
        conn.executemany("INSERT INTO param_T0(name,a,e,i,longnode,argperi,meananom) VALUES (?,?,?,?,?,?,?)", params)
        del params
        T = time.time()-ti; h=int(T/3600); rest=T%3600; m = int(rest/60); s=rest%60
        if verbose: print('  Done in Time: %02i : %02i : %05.2f' %(h,m,s))
    
        # Positions at reference epoch
        if verbose: print('* Calculating positions at reference epoch')
        ti = time.time()
        states0 = self.get_epoch_state(t0, ast_indexs=ast_indexs)#, njobs=njobs)
        T = time.time()-ti; h=int(T/3600); rest=T%3600; m = int(rest/60); s=rest%60
        print('  Done in Time: %02i : %02i : %05.2f' %(h,m,s))
        
        ## Reference epoch
        if verbose: print('* Saving states at reference epoch')
        ti = time.time()
        ### Create state table
        table_name = "state_T0"
        conn.execute("CREATE TABLE state_T0(name TEXT, x REAL,y REAL,z REAL,vx REAL,vy REAL,vz REAL, PRIMARY KEY(name))")
        ### Fill times table
        conn.execute("INSERT INTO state_times(table_title, time) VALUES (?, ?)", (table_name,t0))
        ### Fill state table
        conn.executemany("INSERT INTO state_T0(name,x,y,z,vx,vy,vz) VALUES (?,?,?,?,?,?,?)", np.hstack((pkey, states0)))
        T = time.time()-ti; h=int(T/3600); rest=T%3600; m = int(rest/60); s=rest%60
        if verbose: print('  Done in Time: %02i : %02i : %05.2f' %(h,m,s))

        # Closing db
        conn.commit()
        conn.close()

    def dbconn(self, sqlfile=PACKAGE_PATH+'/aleph_states.db'):
        """
        Returns an SQL connection (sqlite3.connect) to the database 
        (see 'sqlite3' documentation for all options). Remember to
        close the connection when finished or before using any 'aleph'
        function.
        """
        return sqlite3.connect(sqlfile)
        
    def adding_states_into_db (self, tdb_jds, sqlfile=PACKAGE_PATH+'/aleph_states.db', njobs=1, onebodysim=True, verbose=True):
        """
        Integrates states of all bodies in 'sqlfile' in the given 'tdb_jds' epochs
        starting from the states already stored in the database.

        Integration is made using `rebound.Simulation`, wich receives the Sun, all planets
        (and Pluto) and one or more asteroids (determined by 'onebodysim' parameter) starting
        from the states at a near time stored in the database ('sqlfile' parameter).
        
        Parameters:
                    tdb_jds: Array with the requested epochs in JD format and TDB scale.
                    sqlfile: File name of SQL database. Default 'aleph_states.db'.
                    njobs: int Number of parallel processes to run calculations. Default 1.
                    verbose: bool. If True, print messages indicating progress. Default True.
                    onebodysim: bool. If True, each `rebound.Simulation` receives only one asteroid.
                                If False, each `rebound.Simulation` receives several asteroids
                                (determined by 'njobs').
        
        Setting 'onebodysim=True' occupies much more memory (required for all the planets in each
        simulation) but the simulations can run much faster. Setting 'onebodysim=False' saves much
        more memory, but if there are close encounters or other fenomena that could produce an 
        intensive integration (that affects not only to the asteroid that suffers it, but also all
        the other bodies in the simulation, that can be hundreds of thousands) the simulation can
        take an extremly long time to run.

        Take into account that you need to have around 300MB to load astorb.dat and around 115MB
        for one state of all asteroids (without taking into account the Sun and Planets' states)
        before you decide wether to set 'onebodysim' to True.
        """
        # Opening SQL DB
        if verbose: print('* Opening SQL Database')
        conn = sqlite3.connect(sqlfile)
        c = conn.cursor()
        
        # Checking validity of times
        c.execute("SELECT time FROM state_times")
        ts_db = np.array(c.fetchall()).ravel()
        t_add = list(set(tdb_jds)-set(ts_db))
        if len(t_add)==0: 
            warnings.warn("All requested epochs are in the database. Ending task.")
            return
        
        # Getting nearest time in database
        if verbose: print('* Finding starting time in DB')
        min_t = min(t_add); max_t = max(t_add)
        c.execute("SELECT table_title, time, abs(time-?) FROM state_times ORDER BY ABS(time-?)", (min_t,min_t))
        (table_min, min_t_db, min_t_diff) = c.fetchone()
        c.execute("SELECT table_title, time, abs(time-?) FROM state_times ORDER BY ABS(time-?)", (max_t,max_t))
        (table_max, max_t_db, max_t_diff) = c.fetchone()
        if min_t_diff <= max_t_diff: 
            t_start=min_t_db; dt = .1
        else:
            t_start=max_t_db; dt = -.1
    
        # Getting states at nearest time
        if verbose: print('* Getting states at starting time in DB')
        c.execute("SELECT table_title FROM state_times WHERE time=?", (t_start,))
        (table,) = c.fetchone()
        df = pd.read_sql('SELECT * FROM %s' %table, conn, index_col='name')
        pkey = np.reshape(df.index.values, (len(df),1))
        states0 = df.values
        ss_states = get_SS_states(Time(t_start, format='jd', scale='tdb'), frame='helio')

        # Integrating from starting states to requested times
        if dt>0: delta_tdbs = np.array(sorted(t_add)) - t_start
        else: delta_tdbs = np.array(sorted(t_add, reverse=True)) - t_start
        if njobs == 1:
            if verbose: print('* Integrating (can take several hours)')
            ti_ = time.time()
            if onebodysim:
                results = []
                for i in range(len(pkey)):
                    results.append(get_integration([states0[i]], ss_states, dt, delta_tdbs))
                T = time.time()-ti_; h_=int(T/3600); rest=T%3600; m_ = int(rest/60); s_=rest%60
                if verbose: print('    Done in Time: %02i : %02i : %05.2f' %(h_,m_,s_))
                if verbose: print('*   Joining results')
                dicc_integ = {}
                for dt in delta_tdbs:
                    sts = [dicc[dt] for dicc in results]
                    dicc_integ[dt] = np.vstack(sts)
                T = time.time()-ti_; h_=int(T/3600); rest=T%3600; m_ = int(rest/60); s_=rest%60
                if verbose: print('        Done in Time: %02i : %02i : %05.2f' %(h_,m_,s_))
            else:
                dicc_integ = get_integration(states0, ss_states, dt, delta_tdbs)
                T = time.time()-ti_; h_=int(T/3600); rest=T%3600; m_ = int(rest/60); s_=rest%60
                if verbose: print('    Done in Time: %02i : %02i : %05.2f' %(h_,m_,s_))
        else:
            if verbose: print('* Integrating in parallel (can take several hours)')
            ti_ = time.time()
            l = len(pkey); intervalo=l*1./njobs
            argums = []
            if onebodysim:
                for i in range(l):
                    argums.append(([states0[i]], ss_states, dt, delta_tdbs))
            else:
                for i in range(njobs):
                    i_0 = int(i*intervalo); i_f = int((i+1)*intervalo)
                    if i==njobs-1: i_f = l
                    argums.append((states0[i_0:i_f], ss_states, dt, delta_tdbs))
            pool = mp.Pool(processes=njobs)
            ##results = [pool.apply(get_integration, args=arg) for arg in argums]
            ##res_raw = []
            ##for p in results:
            ##    r = p.get(); res_raw.append(r)
            results = pool.starmap(get_integration, argums)
            pool.close()
            T = time.time()-ti_; h_=int(T/3600); rest=T%3600; m_ = int(rest/60); s_=rest%60
            if verbose: print('    Done in Time: %02i : %02i : %05.2f' %(h_,m_,s_))
            if verbose: print('*   Joining parallel results')
            ti_ = time.time()
            dicc_integ = {}
            for dt in delta_tdbs:
                sts = [dicc[dt] for dicc in results]
                dicc_integ[dt] = np.vstack(sts)
            T = time.time()-ti_; h_=int(T/3600); rest=T%3600; m_ = int(rest/60); s_=rest%60
            if verbose: print('        Done in Time: %02i : %02i : %05.2f' %(h_,m_,s_))
    
        # Saving Integration
        if verbose: print('* Saving states')
        ti_ = time.time()
        lt = len(ts_db)+1
        for j, it in enumerate(delta_tdbs):
            tdbs = t_start+it; table_name = "state_T%04i" %(j+lt)
            #if verbose: print('        ', j+lt, it, table_name, tdbs)
            data = np.hstack((pkey, dicc_integ[it]))
            ## Fill times table
            conn.execute("insert into state_times(table_title, time) values (?, ?)", (table_name,tdbs))
            ## Create state table    
            conn.execute("create table %s(name TEXT, x REAL,y REAL,z REAL,vx REAL,vy REAL,vz REAL, primary key(name))" %table_name)
            ## Fill state table
            conn.executemany("insert into "+table_name+"(name, x,y,z,vx,vy,vz) values (?,?,?,?,?,?,?)", data)
        T = time.time()-ti_; h_=int(T/3600); rest=T%3600; m_ = int(rest/60); s_=rest%60
        if verbose: print('    Done in Time: %02i : %02i : %05.2f' %(h_,m_,s_))
        del dicc_integ; del data

        # Closing db
        conn.commit()
        conn.close()

    def adding_params_into_db (self, tdb_jds=None, sqlfile=PACKAGE_PATH+'/aleph_states.db', verbose=True):
        """
        Function that receives epochs with states already stored in the database and computes their
        corresponding orbital parameters (using the 2-body equations). Take into considerations that
        the states are computed in the heliocentric equatorial frame while the orbital parameters
        are in the heliocentric ecliptic frame (as those in astorb.dat or MPCORB.DAT).
        
        Parameters:
                    tdb_jds: Array with the requested epochs in JD format and TDB scale.
                             These epochs must have stored states. Default to None (meaning that
                             orbital parameters will be computed for states at all apochs).
                    sqlfile: File name of SQL database. Optional.
                    verbose: bool. If True, print messages indicating progress. Default True.
        """
        ttotal_ = time.time()
        # Opening SQL DB
        if verbose: print('* Opening SQL Database')
        conn = sqlite3.connect(sqlfile)
        c = conn.cursor()

        c.execute("SELECT time FROM state_times")
        ts_db = np.array(c.fetchall()).ravel()
        
        if tdb_jds is None: tdb_jds = ts_db
        
        # Checking validity of times
        t_intersec = set(tdb_jds)&set(ts_db)
        if len(t_intersec)==0:
            warnings.warn("None of the requested epochs have states calculated. Ending task.")
            return
        c.execute("SELECT time FROM param_times")
        ts_db = np.array(c.fetchall()).ravel()
        t_add = list(set(tdb_jds)-set(ts_db))
        if len(t_add)==0: 
            warnings.warn("All requested epochs have orbital parameters calculated. Ending task.")
            return

        # Iterating for each epoch
        for t in t_add:
            # Computing orbital parameters
            if verbose: print('    Calculating orbital parameters for epoch =',t)
            ti_ = time.time()
            c.execute("SELECT table_title FROM state_times WHERE time=?", (t,))
            (table,) = c.fetchone()
            
            df = pd.read_sql('SELECT * FROM %s' %table, conn, index_col='name')
            pkey = np.reshape(df.index.values, (len(df),1))
            states = df.values
            
            params = []
            for state in states: params.append( pycutils.orbparams_from_state(*state) )
            T = time.time()-ti_#; h_=int(T/3600); rest=T%3600; m_ = int(rest/60); s_=rest%60
            if verbose: print('       Done in %05.2f s' %(T))
            
            # Saving orbital parameters
            if verbose: print('    Saving orbital parameters for epoch =',t)
            ti_ = time.time()
            ## Getting table name
            table_name = 'param_'+table.split('_')[1]
            ## Fill times table
            conn.execute("INSERT INTO param_times(table_title, time) VALUES (?, ?)", (table_name,t))
            ## Create state table
            data = np.hstack((pkey, params))
            conn.execute("create table %s(name TEXT,a REAL,e REAL,i REAL,longnode REAL,argperi REAL,meananom REAL, PRIMARY KEY(name))" %table_name)
            ## Fill state table
            conn.executemany("insert into "+table_name+"(name,a,e,i,longnode,argperi,meananom) values (?,?,?,?,?,?,?)", data)
            for state in states: params.append( pycutils.orbparams_from_state(*state) )
            T = time.time()-ti_#; h_=int(T/3600); rest=T%3600; m_ = int(rest/60); s_=rest%60
            if verbose: print('       Done in %05.2f s' %(T))

        # Closing db
        conn.commit()
        conn.close()

        T = time.time()-ttotal_; h_=int(T/3600); rest=T%3600; m_ = int(rest/60); s_=rest%60
        if verbose: print('    Done in Time: %02i : %02i : %05.2f' %(h_,m_,s_))

    def get_available_db_times(self, sqlfile=PACKAGE_PATH+'/aleph_states.db'):
        """
        Returns an `astropy.time.Time` object with all available times stored in the database. 
        """
        conn = sqlite3.connect(sqlfile)
        c = conn.cursor()
        c.execute("SELECT time FROM state_times")
        ts = np.array(c.fetchall()).ravel()
        conn.close()
        return Time(ts, format='jd', scale='tdb')



import ctypes
import os

# Find .so full suffix
import sysconfig
suffix = sysconfig.get_config_var('EXT_SUFFIX')
if suffix is None:
    suffix = ".so"


PACKAGE_PATH = os.path.abspath(os.path.dirname(__file__))

cutils = ctypes.CDLL(PACKAGE_PATH+"/../cutils"+suffix)

cutils.state_equatorial_heliocentric.argtypes = (ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_double,
                                                 ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_double,
                                                 ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double),
                                                 ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double))

cutils.apparent_coords_equatorial_heliocentric.argtypes = (ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_double,
                                                           ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_double,
                                                           ctypes.c_double, ctypes.c_double, ctypes.c_double,
                                                           ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double),
                                                           ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double),
                                                           ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double))

cutils.params_from_state.argtypes = (ctypes.c_double, ctypes.c_double, ctypes.c_double,
                                     ctypes.c_double, ctypes.c_double, ctypes.c_double,
                                     ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double),
                                     ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double))

def state_equatorial_heliocentric(epoch, paramsepoch, a, e, i, longnode, argperi, meananom):
    """
    Given an epoch and orbital parameters (in heliocentric ecliptic coordinates)
    returns the state (in heliocentric equatorial coordinates).

    Parameters:
               epoch : float. Epoch where the state is calculated (in days).
               paramsepoch: float. Epoch of the orbital parameters (in days).
               a : float. Semi-major axis (in au).
               e : float. Eccentricity.
               i : float. Inclination (in degrees).
               longnode: float. Longitude of ascending node (in degrees).
               argperi: float. Argument of perihelion (in degrees).
               meananom: floaf. Mean anomaly (at pamsepoch, in degrees).
    
    Returns: tuple of floats: x, y, z, vx, vy, vz. State in au and au/day
    """
    jd_ = ctypes.c_double(epoch); tjdepoch_ = ctypes.c_double(paramsepoch)
    
    a_ = ctypes.c_double(a); e_ = ctypes.c_double(e); i_ = ctypes.c_double(i)
    longnode_ = ctypes.c_double(longnode)
    argperi_ = ctypes.c_double(argperi)
    meananom_ = ctypes.c_double(meananom)

    x = ctypes.c_double(0); y = ctypes.c_double(0); z = ctypes.c_double(0)
    vx = ctypes.c_double(0); vy = ctypes.c_double(0); vz = ctypes.c_double(0)
    
    args = (jd_, tjdepoch_, a_, e_, i_, longnode_, argperi_, meananom_,
            x, y, z, vx, vy, vz)

    cutils.state_equatorial_heliocentric(*args)
    return x.value, y.value, z.value, vx.value, vy.value, vz.value

def apparent_coords_equatorial_heliocentric(epoch, obsx, obsy, obsz, paramsepoch, a, e, i, longnode, argperi, meananom):
    """
    Given an epoch, orbital parameters (in heliocentric ecliptic coordinates)
    and observer's coordinates (in heliocentric equatorial coordinates)
    returns the 'apparent' (as seen by the observer, namely, corrected by 
    light-travel time) coordinates, state and distance.

    Parameters:
               epoch : float. Epoch where the state is calculated (in days).
               obsx : float. Observer's 'x' coordinate (in au).
               obsy : float. Observer's 'y' coordinate (in au).
               obsz : float. Observer's 'z' coordinate (in au).
               paramsepoch: float. Epoch of the orbital parameters (in days).
               a : float. Semi-major axis (in au).
               e : float. Eccentricity.
               i : float. Inclination (in degrees).
               longnode: float. Longitude of ascending node (in degrees).
               argperi: float. Argument of perihelion (in degrees).
               meananom: floaf. Mean anomaly (at pamsepoch, in degrees).
    
    Returns: tuple of floats: ra, dec, x, y, z, vx, vy, vz, dist.
             ra, dec : Equatorial coordinates of body as seen by the observer.
             x, y, z, vx, vy, vz : Heliocentric state of body in au and au/day
                                   when observed light was emited.
             dist : Apparent distance observer-body in au. 
                    Equivalent to sqrt(x*x+y*y+z*z). 
    """
    jd_ = ctypes.c_double(epoch); tjdepoch_ = ctypes.c_double(paramsepoch)
    
    xobs = ctypes.c_double(obsx)
    yobs = ctypes.c_double(obsy); zobs = ctypes.c_double(obsz)
    
    a_ = ctypes.c_double(a); e_ = ctypes.c_double(e); i_ = ctypes.c_double(i)
    longnode_ = ctypes.c_double(longnode)
    argperi_ = ctypes.c_double(argperi)
    meananom_ = ctypes.c_double(meananom)

    ra = ctypes.c_double(0); dec = ctypes.c_double(0)
    x = ctypes.c_double(0); y = ctypes.c_double(0); z = ctypes.c_double(0)
    vx = ctypes.c_double(0); vy = ctypes.c_double(0); vz = ctypes.c_double(0)
    gdist = ctypes.c_double(0)
    
    args = (jd_, xobs, yobs, zobs, tjdepoch_, a_, e_, i_, longnode_, argperi_, meananom_,
            ra, dec, x, y, z, vx, vy, vz, gdist)

    cutils.apparent_coords_equatorial_heliocentric(*args)
    return ra.value, dec.value, x.value, y.value, z.value, vx.value, vy.value, vz.value, gdist.value

def orbparams_from_state(x, y, z, vx, vy, vz):
    """
    Given a body's state (in heliocentric equatorial coordinates), returns 
    orbital parameters (in heliocentric ecliptic coordinates).

    Parameters:
               x : float. Body's 'x' coordinate (in au).
               y : float. Body's 'y' coordinate (in au).
               z : float. Body's 'z' coordinate (in au).
               vx : float. Body's 'vx' velocity component (in au/day).
               vy : float. Body's 'vy' velocity component (in au/day).
               vz : float. Body's 'vz' velocity component (in au/day).
    
    Returns: tuple of floats: a, e, i, longnode, z, argperi, meananom.
             a : Semi-major axis (in au).
             e : Eccentricity.
             i : Inclination (in degrees).
             longnode: Longitude of ascending node (in degrees).
             argperi: Argument of perihelion (in degrees).
             meananom: Mean anomaly (at pams. epoch, in degrees).
    """
    x_ = ctypes.c_double(x); y_ = ctypes.c_double(y); z_ = ctypes.c_double(z)
    vx_ = ctypes.c_double(vx); vy_ = ctypes.c_double(vy); vz_ = ctypes.c_double(vz)
    
    a = ctypes.c_double(0); e = ctypes.c_double(0); i = ctypes.c_double(0)
    longnode = ctypes.c_double(0); argperi = ctypes.c_double(0)
    meananom = ctypes.c_double(0)
    
    args = (x_, y_, z_, vx_, vy_, vz_, a, e, i, longnode, argperi, meananom)

    cutils.params_from_state(*args)
    return a.value, e.value, i.value, longnode.value, argperi.value, meananom.value

PI         = 3.14159265358979323846
TWOPI      = 6.283185307179586476925287
#C          = 173.14463348
G0         = 39.476926421373015
Msun       = 1.0000059769998062e+00
SSMASS     = 1.00134       # Total SS mass
DAY        = (1./365.25)	# Julian day, 86400 s
AXIAL_TILT = (23+26/60.+21.406/3600.)*PI/180. # tilt between equatorial and ecliptic coordinates ec=rot_x(eq,A_T); eq=rot_x(ec,-A_T)
ASEC2RAD   = 4.848136811095359935899141e-6 # Constant conversion from arcsec to rad. From NOVAS.
C_AUDAY    = 173.1446326846693



from setuptools import setup, Extension

# Getting Description
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

# C Extension
cutilsmodule = Extension('cutils',
                    sources = [ 'aleph/cutils.c',],
                    include_dirs = ['aleph'],
                    #define_macros=[ ('CUTILS', None) ],
                    # Removed '-march=native' for now.
                    extra_compile_args=['-O3','-Wno-unknown-pragmas', '-fPIC', '-shared'], # '-DCUTILS', ghash_arg, '-fstrict-aliasing', '-std=c99', '-D_GNU_SOURCE'
                    #extra_link_args=extra_link_args,
                    language = 'c',
                    )

setup(
      name="astro_aleph", #"aleph",
      version="1.0.0",
      author="Jose Pena Z.",
      author_email="jpena@das.uchile.cl",
      description="ALeRCE's Ephemeris",
      long_description=long_description,
      long_description_content_type="text/markdown",
      url="https://github.com/josepenaz/ephs",
      #project_urls={
      #    "Bug Tracker": "https://github.com/pypa/sampleproject/issues",
      #},
      #classifiers=[
      #    "Programming Language :: Python :: 3",
      #    "License :: OSI Approved :: MIT License",
      #    "Operating System :: OS Independent",
      #],
      package_dir={"aleph": "aleph"},
      packages=['aleph'],
      python_requires=">=3.6",
      install_requires=['numpy>=1.18.1',
                        'astropy>=4.2',
                        'pandas>=1.1.3',
                        'rebound>=3.12.2',
                        'jplephem'
                        ],
      package_data = {'aleph':['cutils.h']},
      ext_modules = [cutilsmodule],

)
